"""Resolwe annotations serializer."""

from typing import Any

from django.db import models, transaction
from rest_framework import serializers
from rest_framework.fields import empty

from resolwe.flow.models.annotations import NAME_LENGTH as ANNOTATION_NAME_LENGTH
from resolwe.flow.models.annotations import (
    AnnotationField,
    AnnotationGroup,
    AnnotationPreset,
    AnnotationValue,
)

from .base import ResolweBaseSerializer
from .contributor import ContributorSerializer
from .fields import DictRelatedField, PrimaryKeyDictRelatedField


class AnnotationFieldListSerializer(serializers.ListSerializer):
    """Override list serializer to allow filtering nested relation."""

    def to_representation(self, data: "models.QuerySet[AnnotationGroup]"):
        """Filter the annotation fields within the group by the collection."""
        if (
            "request" in self.context
            and "collection_id" in self.context["request"].query_params
        ):
            collection_id = self.context["request"].query_params["collection_id"]
            data = data.filter(collections__id=collection_id)
        return super().to_representation(data)


class AnnotationGroupSerializer(ResolweBaseSerializer):
    """Serializer for AnnotationGroup objects."""

    class Meta:
        """AnnotationGroupSerializer Meta options."""

        model = AnnotationGroup
        read_only_fields = ("id", "label", "name", "sort_order")
        fields = read_only_fields


class AnnotationFieldSerializer(ResolweBaseSerializer):
    """Serializer for AnnotationField objects."""

    group = DictRelatedField(
        queryset=AnnotationGroup.objects.all(), serializer=AnnotationGroupSerializer
    )

    class Meta:
        """AnnotationFieldSerializer Meta options."""

        model = AnnotationField
        read_only_fields = ("id",)
        fields = read_only_fields + (
            "description",
            "group",
            "label",
            "name",
            "sort_order",
            "type",
            "validator_regex",
            "vocabulary",
            "required",
            "version",
        )
        list_serializer_class = AnnotationFieldListSerializer


class AnnotationFieldDictSerializer(serializers.Serializer):
    """Serializer for actions that deal with annotation fields."""

    annotation_fields = PrimaryKeyDictRelatedField(
        queryset=AnnotationField.objects.all(), many=True
    )


class AnnotationsByPathSerializer(serializers.Serializer):
    """Serializer that reads annotation field and its value."""

    # The field path contains the annotation group name and the annotation field name
    # separated by spaces.
    field_path = serializers.CharField(max_length=2 * ANNOTATION_NAME_LENGTH + 1)
    value = serializers.JSONField(allow_null=True)


class AnnotationPresetSerializer(ResolweBaseSerializer):
    """Serializer for AnnotationPreset objects."""

    class Meta:
        """AnnotationPresetSerializer Meta options."""

        model = AnnotationPreset
        read_only_fields = ("id",)
        fields = read_only_fields + ("name", "fields", "contributor")


class AnnotationValueListSerializer(serializers.ListSerializer):
    """Perform bulk update of annotation values to speed up requests."""

    @transaction.atomic
    def create(self, validated_data: Any) -> Any:
        """Perform bulk create."""
        # The save must be called for delete markers to be set correctly.
        values = [AnnotationValue(**data) for data in validated_data]
        for value in values:
            value.save()
        return values

    @transaction.atomic
    def save(self, **kwargs: Any) -> list[AnnotationValue]:
        """Save and return the list of annotation values."""
        for data in self.validated_data:
            data["value"] = data["_value"]["value"]
        self.instance = AnnotationValue.objects.add_annotations(self.validated_data)
        return self.instance

    def validate(self, attrs: Any) -> Any:
        """Validate list of annotation values."""
        if len(set((attr["field"], attr["entity"]) for attr in attrs)) != len(attrs):
            raise serializers.ValidationError(
                "Duplicate annotation values for the same entity and field."
            )
        return super().validate(attrs)


class AnnotationValueSerializer(ResolweBaseSerializer):
    """Serializer for AnnotationValue objects."""

    contributor = ContributorSerializer()

    def __init__(self, instance=None, data=empty, **kwargs):
        """Rewrite value -> _value."""
        if data is not empty:
            for entry in data if isinstance(data, list) else [data]:
                if "value" in entry:
                    entry["_value"] = {"value": entry.pop("value")}
        super().__init__(instance, data, **kwargs)

    def get_unique_together_validators(self) -> list:
        """Return the list of validators for unique together model constraint.

        Since we are using put to create or update existing values this must be empty
        or updating existing values with put method will fail the validation.
        """
        return []

    class Meta:
        """AnnotationValueSerializer Meta options."""

        model = AnnotationValue
        read_only_fields = ("label", "created")
        update_protected_fields = ("id", "entity", "field", "contributor")
        fields = read_only_fields + update_protected_fields + ("value", "_value")
        extra_kwargs = {"_value": {"write_only": True}}
        list_serializer_class = AnnotationValueListSerializer



"""Resource consumption estimators for processes."""

from collections import defaultdict

PROCESS_RESOURCES = defaultdict(dict)


def estimator(process_slug, resource):
    """Register resource consumption estimator."""

    def decorator(func):
        """Register resource consumption estimator."""
        PROCESS_RESOURCES[process_slug][resource] = func
        return func

    return decorator



import os
from importlib import metadata

import django
from django.conf import settings

base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# Manual setup is required for standalone Django usage
# NOTE: Since documentation is built using the built/installed package when
# using Tox, it can't use the 'test.settings' Django settings module.
settings.configure(
    INSTALLED_APPS=(
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.staticfiles",
        "rest_framework",
        "resolwe",
        "resolwe.permissions",
        "resolwe.flow",
    ),
    DEBUG=True,
)
django.setup()

# Get package metadata from 'resolwe/__about__.py' file
about = {}
with open(os.path.join(base_dir, "resolwe", "__about__.py")) as f:
    exec(f.read(), about)

# -- General configuration ------------------------------------------------

# The extension modules to enable.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
]

# The suffix(es) of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "Resolwe"
version = metadata.metadata("resolwe")["version"]
release = version
author = metadata.metadata("resolwe")["Author-email"]
copyright = "2015-2025, " + author

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# Warn about all references where the target cannot be found
nitpicky = True
# Except for the following:
nitpick_ignore = [
    # This is needed to prevent warnings for container types, e.g.:
    # :type foo: tuple(bool, str)
    ("py:obj", ""),
]

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.
html_theme = "sphinx_rtd_theme"

# Output file base name for HTML help builder.
htmlhelp_basename = "Resolwedoc"

# Configuration for intersphinx
_django_major_version = "{}.{}".format(*django.VERSION[:2])
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "django": (
        "https://docs.djangoproject.com/en/{}/".format(_django_major_version),
        "https://docs.djangoproject.com/en/{}/_objects/".format(_django_major_version),
    ),
}

# Configuration for extlinks
extlinks = {
    "drf": ("http://www.django-rest-framework.org/api-guide/%s", ""),
}



""".. Ignore pydocstyle D400.

========
Listener
========

Standalone client used as a contact point for executors. It uses zeromq
socket for communication.

.. autoclass:: resolwe.flow.managers.listener.ExecutorListener
    :members:

"""

import asyncio
import datetime
import logging
import os
from contextlib import suppress
from functools import lru_cache
from time import time
from typing import Any, ChainMap, Dict, List, Optional, Set, Union

import zmq
import zmq.asyncio
from asgiref.sync import async_to_sync
from channels.db import database_sync_to_async
from channels.layers import get_channel_layer
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import transaction
from django.utils.timezone import now

from resolwe.flow.executors.socket_utils import (
    BaseProtocol,
    Message,
    MessageProcessingEventType,
    PeerIdentity,
    Response,
    ResponseStatus,
)
from resolwe.flow.executors.zeromq_utils import ZMQCommunicator
from resolwe.flow.managers import consumer
from resolwe.flow.managers.listener.metrics import metrics_reporter
from resolwe.flow.managers.protocol import WorkerProtocol
from resolwe.flow.managers.state import LISTENER_CONTROL_CHANNEL  # noqa: F401
from resolwe.flow.models import Data, Storage, Worker
from resolwe.flow.utils import iterate_schema
from resolwe.storage.models import AccessLog
from resolwe.test.utils import ignore_in_tests, is_testing
from resolwe.utils import BraceMessage as __

from .authenticator import ZMQAuthenticator

# Register plugins by importing them.
from .basic_commands_plugin import BasicCommands  # noqa: F401
from .bootstrap_plugin import BootstrapCommands  # noqa: F401
from .plugin import listener_plugin_manager as plugin_manager
from .python_process_plugin import PythonProcess  # noqa: F401
from .redis_cache import RedisLockStatus, cache_manager, redis_server

# Unique redis object to use in listener.

logger = logging.getLogger(__name__)
User = get_user_model()

# Public and private key read from the environment. When the environment does not
# exist (for instance when running tests), default key is used.
# The default key can only be used when the DEBUG setting is set to True.
DEFAULT_LISTENER_PUBLIC_KEY = "*%z%0!H7z84uT^]Gwz89:YQr#@Wm=qB(9Ao+qQ^K"
DEFAULT_LISTENER_PRIVATE_KEY = ":M-RdvbLHztG+FaaYdp}RKV-fajY}$QT[+XgWD{="
env_public_key = os.getenv("LISTENER_PUBLIC_KEY", DEFAULT_LISTENER_PUBLIC_KEY)
env_private_key = os.getenv("LISTENER_PRIVATE_KEY", DEFAULT_LISTENER_PRIVATE_KEY)

if not settings.DEBUG:
    assert (
        env_private_key != DEFAULT_LISTENER_PRIVATE_KEY
        and env_public_key != DEFAULT_LISTENER_PUBLIC_KEY
    ), "Must set private and public listener key when not running in debug mode"

# Make sure listener private and public keys are encoded.
LISTENER_PUBLIC_KEY = env_public_key.encode()
LISTENER_PRIVATE_KEY = env_private_key.encode()

if not settings.DEBUG:
    assert (
        LISTENER_PRIVATE_KEY != DEFAULT_LISTENER_PRIVATE_KEY
        and LISTENER_PUBLIC_KEY != DEFAULT_LISTENER_PUBLIC_KEY
    ), "Must set private and public listener key when not running in debug mode"


class Processor:
    """Process the messages sent by the workers.

    Some of the data attributes are cached by redis.

    Important: the consistincy of the cache is guaranteed only when the
    `_update_data`  or `_save_data` method is used to save data.
    """

    def __init__(self, listener: "ListenerProtocol"):
        """Initialize variables."""
        self._listener = listener
        self._return_codes: Dict[PeerIdentity, int] = dict()

    def get_data_fields(self, data_id: int, field_names: Union[List[str], str]) -> Any:
        """Get the data fields for the data with the given id.

        Note: no permission check is done so only call this method on Data
            objects user is clear to access.

        :attr data_id: the id of the data object.
        :attr field_names: the names of the fields. This can be a sequence or
            a string.
        :return: a single value when field_names is a string or a tuple of
            values when it is a sequence of strings.
        """
        identifiers = (data_id,)

        fields = [field_names] if isinstance(field_names, str) else field_names
        # Retrieve the cached data. If no data is cached the result is empty dict.
        cached_data = cache_manager.get(Data, identifiers) or {}
        # Retrieve the missing fields from the database.
        database_fields = {field for field in fields if field not in cached_data}

        # Make sure to only read the data if database fields are not empty.
        # Django interprets empty dict as "read all fields".
        database_data = {}
        if database_fields:
            database_data = (
                Data.objects.filter(pk=data_id).values(*database_fields).get()
            )
        if database_data:
            cache_manager.update_cache(Data, (data_id,), database_data)
        # Return the combined results, with database data taking precedence.
        combined = ChainMap(database_data, cached_data)
        result = [combined[field_name] for field_name in fields]
        return result[0] if isinstance(field_names, str) else result

    @ignore_in_tests(lru_cache(maxsize=100))
    def contributor_id(self, data_id: int) -> int:
        """Get the id of the user that created the given data object.

        This function is cached since contributor is immutable, except in tests,
        where ids may be reused.
        """
        return self.contributor(data_id).id

    @ignore_in_tests(lru_cache(maxsize=100))
    def contributor(self, data_id: int):
        """Get the user that created the given data objects.

        This function is cached since contributor is immutable, except in tests,
        where ids may be reused.
        """
        return User.objects.get(data__id=data_id)

    def data(self, data_id: int) -> Data:
        """Get the data object for the given data id.

        Use with caution: loading the entire data object is slow.
        """
        return Data.objects.get(pk=data_id)

    def worker(self, data_id: int) -> Worker:
        """Get the worker object.

        The evaluation is done when needed in order not to make potentially
        failing SQL query in the constructor.

        This method uses Django ORM.
        """
        return Worker.objects.get(data=data_id)

    def storage_fields(self, data_id) -> Set[str]:
        """Get the names of storage fields in the schema.

        :attrs data_id: the id of the data object to get the fields for.
        """
        output_schema = self.get_data_fields(data_id, "process__output_schema")
        return {
            field_name
            for schema, _, field_name in iterate_schema({}, output_schema)
            if schema["type"].startswith("basic:json:")
        }

    def save_storage(self, key: str, content: str, data: Data) -> Storage:
        """Save storage field and add it to the given data.

        :return: the pk of the saved storage object.
        :raise TypeError: when the data object does not exist.
        """

        storage_pk = data.output.get(key)
        storage_defaults = {
            "json": content,
            "name": f"Storage for data id {data.pk}",
            "contributor_id": data.contributor_id,
        }
        storage: Storage = Storage.objects.update_or_create(
            pk=storage_pk, defaults=storage_defaults
        )[0]
        storage.data.add(data)
        return storage

    def _unlock_all_inputs(self, data_id: int):
        """Unlock all data objects that were locked by the given data.

        If exception occurs during unlocking we can not do much but ignore it.
        The method is called when dispatcher is nudged.
        """
        with suppress(Exception):
            query = AccessLog.objects.filter(cause_id=data_id)
            query.update(finished=now())

    def _choose_worst_status(self, status1: str, status2: str) -> str:
        """Get more problematic status of the two.

        Assumption: statuses in Data.STATUS_CHOICES are ordered from least
            to most problematic.

        :raises AssertionError: when status1 or status2 do not represent
            possible status of Data object.

        """
        statuses = [status[0] for status in Data.STATUS_CHOICES]
        assert status1 in statuses
        assert status2 in statuses
        return max([status1, status2], key=lambda status: statuses.index(status))

    def _save_error(self, data: Data, error: str):
        """Log error to Data object, ignore possible exceptions.

        The status of Data object is changed to ``Data.STATUS_ERROR``.
        """
        logger.debug("Saving error to data object %s", data.id)
        if error not in data.process_error:
            data.process_error.append(error)
        data.status = Data.STATUS_ERROR
        try:
            self._save_data(data, ["process_error", "status"])
        except Exception:
            logger.exception("Error when saving error to the data object.")

    def _log_exception(
        self,
        data: Data,
        error: str,
        extra: Optional[dict] = None,
        save_to_data_object: bool = True,
    ):
        """Log current exception and optionally store it to the data object.

        When argument ``save_to_data_object`` is false the status of the
        data object is not changed.
        """
        if extra is None:
            extra = dict()
        extra.update({"data_id": data.id})
        logger.exception(error, extra=extra)
        if save_to_data_object:
            self._save_error(data, error)

    def _log_error(
        self,
        data: Data,
        error: str,
        extra: dict = {},
        save_to_data_object: bool = True,
    ):
        """Log error and optionally store it to the data object.

        When argument ``save_to_data_object`` is false the status of the
        data object is not changed.
        """
        extra.update({"data_id": data.id})
        logger.error(error, extra=extra)
        if save_to_data_object:
            assert data is not None, "Can not save error to None object."
            self._save_error(data, error)

    def _save_data(self, data: Data, changes: Union[list[str], dict[str, Any]]):
        """Update the data object with the given id.

        :attr changes: dictionary with changes to be saved or a list of fields that
            were modified.
        :attr data_id: the given data id.

        :raises: exception when data object cannot be saved.
        """
        if isinstance(changes, list):
            changes_dict = {field: getattr(data, field) for field in changes}
            update_fields = changes
        else:
            changes_dict = changes
            update_fields = changes.keys()
            for attribute, value in changes_dict.items():
                setattr(data, attribute, value)
        cache_manager.update_cache(Data, (data.id,), changes_dict)
        data.save(update_fields=update_fields)

    def _update_data(self, data_id: int, changes: Dict[str, Any]):
        """Update the data object with the given id.

        Changes are dictionary mapping from field names to field values.

        :attr changes: dictionary with changes to be saved.
        :attr data_id: the given data id.

        :raises: exception when data object cannot be saved.
        """
        if changes:
            # Update the redis cache.
            cache_manager.update_cache(Data, (data_id,), changes)
            Data.objects.filter(pk=data_id).update(**changes)

    def _update_worker(self, data_id: int, changes: Dict[str, Any]):
        """Update the worker object for the given data.

        :raises: exception when data object cannot be saved.
        """
        Worker.objects.filter(data__pk=data_id).update(**changes)
        cache_manager.update_cache(
            Data,
            (data_id,),
            {f"worker__{key}": value for key, value in changes.items()},
        )

    def _save_database_terminate(self, data_id: int):
        """Save error to the database."""
        data = self.data(data_id)
        self._save_error(data, "Processing was cancelled.")

    async def terminate(self, peer_identity: PeerIdentity):
        """Send the terminate command to the worker for the given data object.

        Peer should terminate by itself and send finish message back to us.
        """
        try:
            await database_sync_to_async(
                self._save_database_terminate, thread_sensitive=False
            )(int(peer_identity))
            logger.debug("Sending terminate command to the peer '%s'.", peer_identity)
            await self._listener.communicator.send_command(
                Message.command("terminate", "Terminate worker"),
                peer_identity=peer_identity,
                timeout=5,
            )
            logger.debug("Terminate command to the peer '%s' sent.", peer_identity)
        except Exception:
            logger.exception("Error terminating worker.")

    async def peer_not_responding(self, data_id: int):
        """Peer is not responding, abort the processing.

        TODO: what to do with unresponsive peer which wakes up later and
        starts sending commands to the listener? Currently the situation
        is handled by sending "terminate" command to it and ignore its
        commands.
        """

        def update_database():
            """Update the database status."""
            with transaction.atomic():
                error_message = "Processing task is not responding."
                data = self.data(data_id)
                self._log_error(data, error_message)
                self._save_data(data, ["process_error", "status"])
                self._update_worker(data_id, {"status": Worker.STATUS_NONRESPONDING})

        logger.debug(__("Peer with id={} is not responding.", data_id))
        await database_sync_to_async(update_database, thread_sensitive=False)()
        await self.notify_dispatcher_abort_async(data_id)

    def _can_process_object(
        self, worker_status: str, data_status: str, command_name: str
    ) -> bool:
        """Check if the given data and worker status are ok to process.

        When data is in Data.STATUS_ERROR some commands are still allowed:
        - referenced_files: to store a list of filest to a data object.
        - finish: to set a return code and finalize the data object.
        -
        """
        acceptable_worker_statuses = (
            Worker.STATUS_PROCESSING,
            Worker.STATUS_PREPARING,
            Worker.STATUS_FINISHED_PREPARING,
        )
        unacceptable_data_statuses = (
            Data.STATUS_DONE,
            Data.STATUS_DIRTY,
            Data.STATUS_ERROR,
        )

        if worker_status not in acceptable_worker_statuses:
            return False

        allowed_commands_error = ["referenced_files", "finish"]
        if data_status == Data.STATUS_ERROR and command_name in allowed_commands_error:
            return True

        return data_status not in unacceptable_data_statuses

    def process_command(self, identity: PeerIdentity, message: Message) -> Response:
        """Process a single command from the peer.

        This command is run in the database_sync_to_async so it is safe to
        perform Django ORM operations inside.

        All exceptions will be handled and logged inside this method. The error

        """
        data_id = abs(int(identity))
        # Do not proccess messages from Workers that have already finish
        # processing data objects.
        worker_status, data_status, started = self.get_data_fields(
            data_id, ["worker__status", "status", "started"]
        )
        if not self._can_process_object(
            worker_status, data_status, message.command_name
        ):
            return message.respond_error(
                f"Unable to process the data object {data_id} with status {data_status}."
            )
        handler_name = f"handle_{message.command_name}"
        logger.debug(__("Message for handler {} received.", handler_name))
        handler = plugin_manager.get_handler(message.command_name)
        if not handler:
            error = f"Unknow command '{message.command_name}'."
            self._log_error(self.data(data_id), error, save_to_data_object=False)
            return message.respond_error(error)

        # Set the data started on the first command.
        if started is None:
            self._save_data(self.data(data_id), {"started": now()})

        try:
            logger.debug(__("Invoking handler {}.", handler_name))
            response = handler(data_id, message, self)
            # Check if data status was changed by the handler.
            if self.get_data_fields(data_id, "status") == Data.STATUS_ERROR:
                response.status = ResponseStatus.ERROR
            return response
        except ValidationError as err:
            error = (
                f"Validation error when running handler {handler_name} for "
                f"Data object with the id {data_id}: {err}."
            )
            self._log_exception(self.data(data_id), error)
            return message.respond_error("Validation error")
        except Exception as err:
            error = (
                f"Exception when running handler {handler_name} for data "
                f"object with the id {data_id}: {err}."
            )
            self._log_exception(self.data(data_id), error)
            return message.respond_error(f"Error in command handler '{handler_name}'.")

    def notify_dispatcher_abort(self, data_id: int):
        """Notify dispatcher that processing was aborted.

        .. IMPORTANT::

            This only makes manager's state consistent and doesn't
            affect Data object in any way.

        :param obj: The Channels message object. Command object format:

            .. code-block:: none

                {
                    'command': 'abort',
                    'data_id': [id of the :class:`~resolwe.flow.models.Data` object
                               this command was triggered by],
                }
        """
        async_to_sync(self.notify_dispatcher_abort_async)(data_id)

    async def notify_dispatcher_abort_async(self, data_id: int):
        """Notify dispatcher that processing was aborted.

        :param obj: The Channels message object. Command object format:

            .. code-block:: none

                {
                    'command': 'abort',
                    'data_id': [id of the :class:`~resolwe.flow.models.Data` object
                               this command was triggered by],
                }
        """
        await database_sync_to_async(self._unlock_all_inputs, thread_sensitive=False)(
            data_id
        )
        await consumer.send_event(
            {
                WorkerProtocol.COMMAND: WorkerProtocol.ABORT,
                WorkerProtocol.DATA_ID: data_id,
                WorkerProtocol.FINISH_COMMUNICATE_EXTRA: {
                    "executor": getattr(settings, "FLOW_EXECUTOR", {}).get(
                        "NAME", "resolwe.flow.executors.local"
                    ),
                },
            }
        )
        logger.debug("notify_dispatcher_abort: consumer event sent")

    def notify_dispatcher_finish(self, data_id: int):
        """Notify dispatcher that the processing is finished.

        See ``notify_dispatcher_abort`` for message format.
        """
        async_to_sync(self.notify_dispatcher_finish_async)(data_id)

    async def notify_dispatcher_finish_async(self, data_id: int):
        """Notify dispatcher that the processing is finished.

        See ``notify_dispatcher_abort`` for message format.
        """
        await database_sync_to_async(self._unlock_all_inputs, thread_sensitive=False)(
            data_id
        )
        await consumer.send_event(
            {
                WorkerProtocol.COMMAND: WorkerProtocol.FINISH,
                WorkerProtocol.DATA_ID: data_id,
                WorkerProtocol.FINISH_COMMUNICATE_EXTRA: {
                    "executor": getattr(settings, "FLOW_EXECUTOR", {}).get(
                        "NAME", "resolwe.flow.executors.local"
                    ),
                },
            }
        )


class ListenerProtocol(BaseProtocol):
    """Listener protocol implementation."""

    def __init__(
        self,
        hosts: List[str],
        port: int,
        protocol: str,
        zmq_socket: Optional[zmq.asyncio.Socket] = None,
        max_concurrent_commands: int = 10,
    ):
        """Initialize."""
        if zmq_socket is None:
            zmq_context: zmq.asyncio.Context = zmq.asyncio.Context.instance()
            auth = ZMQAuthenticator.instance(zmq_context)
            auth.start()
            zmq_socket = zmq_context.socket(zmq.ROUTER)
            # Configure keys and start the server.
            zmq_socket.curve_secretkey = LISTENER_PRIVATE_KEY
            zmq_socket.curve_publickey = LISTENER_PUBLIC_KEY
            zmq_socket.curve_server = True
            zmq_socket.setsockopt(zmq.ROUTER_HANDOVER, 1)
            for host in hosts:
                zmq_socket.bind(f"{protocol}://{host}:{port}")

        super().__init__(
            ZMQCommunicator(zmq_socket, "listener <-> workers", logger),
            logger,
            max_concurrent_commands,
            metrics_reporter,
        )
        self.communicator.heartbeat_handler = self.heartbeat_handler
        self._message_processor = Processor(self)

    async def heartbeat_handler(self, peer_identity: PeerIdentity):
        """Handle the heartbeat messages."""
        # The redis key contains the timestamp when the worker was last seen.
        # When the key does not exist in the redis database create one with the
        # current timestamp.
        try:
            one_hour = 3600
            one_day = 24 * one_hour
            data_id = abs(int(peer_identity))
            redis_key = f"resolwe-worker-{data_id}"
            redis_server.set(redis_key, int(time()))

            # Make sure the expiration time is longer than the highest timeout
            # (currently one week). Otherwise the entry will expire before the timeout
            # is reached.
            redis_server.expire(redis_key, 8 * one_day)
        except Exception:
            logger.exception("Exception in heartbeat handler.")

    async def heartbeat_task(self, check_interval=60):
        """Periodically check workers.

        This task must be cancelled to stop.

        :attr check_interval: how many seconds to wait between two consecutive
            checks.
        """
        while True:
            logger.debug("Checking workers.")
            await self.check_workers()
            await asyncio.sleep(check_interval)

    async def check_workers(self):
        """Check all workers and possibly mark them as stalled."""

        def get_data():
            """Get the data from the database.

            The list of workers in non-final state should not be long and can
            be safely stored in a list.
            """
            return list(
                Worker.objects.exclude(status__in=Worker.FINAL_STATUSES).values_list(
                    "data_id", "status"
                )
            )

        default_timeout = 600
        one_hour = 3600
        one_day = 24 * one_hour

        non_responsive_timeout = {
            Worker.STATUS_FINISHED_PREPARING: 2 * one_hour,
            Worker.STATUS_PREPARING: 7 * one_day,
        }

        current_timestamp = int(time())
        for data_id, worker_status in await database_sync_to_async(
            get_data, thread_sensitive=False
        )():
            redis_key = f"resolwe-worker-{data_id}"
            last_seen = redis_server.get(redis_key)
            if last_seen is None:
                redis_server.set(redis_key, current_timestamp)
                # Make sure the expiration time is longer than the highest timeout
                # (currently one week). Otherwise the entry will expire before the
                # timeout is reached.
                redis_server.expire(redis_key, 8 * one_day)
            else:
                last_seen = int(last_seen)
            without_heartbeat = current_timestamp - (last_seen or current_timestamp)
            logger.debug(
                __(
                    "Worker {} with status {} not seen {} seconds.",
                    data_id,
                    worker_status,
                    without_heartbeat,
                )
            )
            if without_heartbeat > non_responsive_timeout.get(
                worker_status, default_timeout
            ):
                try:
                    logger.info(
                        __(
                            "Worker {} with status {} marked non-responsive.",
                            data_id,
                            worker_status,
                        )
                    )
                    await self._message_processor.peer_not_responding(data_id)
                except Exception:
                    self.logger.exception(
                        "Exception updating unresponsive peer status."
                    )

    def _handle_lock_message_error(
        self, lock_status: RedisLockStatus, received_message: Message
    ) -> Response:
        """Construct and return the message to be sent."""
        message = f"Message {received_message.uuid} "
        match lock_status:
            case RedisLockStatus.OK:
                message += "already processed with OK status."
                response_method = received_message.respond_skip
            case RedisLockStatus.ERROR:
                message += "already processed with error status."
                response_method = received_message.respond_error
            case RedisLockStatus.PROCESSING:
                message += "is processing."
                response_method = received_message.respond_skip
        logger.debug(message)
        return response_method(message)

    async def extend_processing_lock(
        self,
        data_id: int,
        message_uuid: str,
        refresh_interval: int = 3 * 60,
        extend_for: int = 5 * 60,
    ):
        """Extend the processing lock for the given message.

        :attr refresh_interval: extend the key every this many seconds.
        :attr extend_for: set the key TTL to this many seconds.
        """
        while True:
            await asyncio.sleep(refresh_interval)
            try:
                cache_manager.extend_lock(
                    Data, [(data_id, message_uuid)], valid_for=extend_for
                )
            except Exception:
                logger.exception("Error extending lock.")

    async def default_command_handler(
        self, received_message: Message, peer_identity: PeerIdentity
    ) -> Response:
        """Process command.

        Lock the command before the processing starts and unlock it when it finishes.
        """
        response: Optional[Response] = None
        extend_lock_task: Optional[asyncio.Task] = None

        # Handle the liveness probe in a special way (no access control is needed).
        if received_message.command_name == "liveness_probe":
            return received_message.respond("OK")

        try:
            data_id = abs(int(peer_identity))
        except Exception:
            error = f"Unable to parse the identity {peer_identity!r}."
            self.logger.error(error)
            return received_message.respond_error(error)

        try:
            # Check that the client is valid and can process the data object.
            # Before any processing make sure the client is valid and has the permission to
            # process the data object.
            key = received_message.client_id
            authenticator = ZMQAuthenticator.instance()
            if not authenticator.can_access_data(key, data_id):
                error_message = (
                    f"Client with key {key!r} is not allowed to process the data object "
                    f"with id {data_id}."
                )
                self.logger.error(error_message)
                return received_message.respond_error(error_message)
        except Exception:
            error = f"Client {received_message.client_id!r} is not valid."
            return received_message.respond_error(error)

        try:
            # Try to lock the specific message for the data object. If locked it means
            # the message was/is being processed by some other listener.
            # If the listener crashes after successfully obtaining the lock, it will be
            # released automatically after five minutes.
            # If the processing of the message takes longer than 10 minutes, then same
            # message can be processed twice. We can avoid processing the message twice
            # but then all processes that have messages in the listener queue will fail
            # if the listener chashes.
            success, lock_status = cache_manager.lock(
                Data, [(data_id, received_message.uuid)]
            )[0]
            if not success:
                response = self._handle_lock_message_error(
                    lock_status, received_message
                )
            else:
                extend_lock_task = asyncio.create_task(
                    self.extend_processing_lock(data_id, received_message.uuid)
                )
                response = await database_sync_to_async(
                    self._message_processor.process_command, thread_sensitive=False
                )(peer_identity, received_message)
        finally:
            # Stop the extend lock task.
            if extend_lock_task is not None:
                extend_lock_task.cancel()
            # Something crashed during processing and the response was not created.
            # Send a generic error response.
            if response is None:
                response = received_message.respond_error("Error processing message.")
            # Unlock the message.
            if response.response_status == ResponseStatus.ERROR:
                unlock_status = RedisLockStatus.ERROR
            else:
                unlock_status = RedisLockStatus.OK
            cache_manager.unlock(
                Data, [(data_id, received_message.uuid)], status=unlock_status
            )
        self.logger.debug(__("Response time: {}", received_message.time_elapsed()))
        return response

    async def post_finish(self, message: Message, peer_identity: PeerIdentity):
        """Notify dispatcher after finish command was received."""
        try:
            data_id = abs(int(peer_identity))
            with suppress(Data.DoesNotExist):
                await self._message_processor.notify_dispatcher_finish_async(data_id)
        except Exception:
            logger.exception("Error processing post finish command.")

    async def post_init_completed(self, message: Message, peer_identity: PeerIdentity):
        """Sent metrics about preparation phase duration."""
        with suppress(Exception):
            data_id = abs(int(peer_identity))
            started: datetime.datetime = (
                await database_sync_to_async(
                    self._message_processor.get_data_fields,
                    thread_sensitive=False,
                )(data_id, ["started"])
            )[0]
            try:
                metrics_reporter.event(
                    MessageProcessingEventType.PREPARATION_FINISHED,
                    message,
                    peer_identity,
                    started=started.timestamp(),
                )
            except:
                logger.exception("Error processing post init_completed command.")


def handle_exceptions(loop, context):
    """Log uncaught exceptions in asyncio."""
    msg = context.get("exception", context["message"])
    future = context.get("future")
    if future is not None:
        name = future.get_coro().__name__
        logger.error(f"Caught exception from {name}: {msg}")
    else:
        logger.error(f"Caught exception: {msg}")


class ExecutorListener:
    """The contact point implementation for executors."""

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        """Initialize."""
        # Read socket parameters.
        self._hosts = list(
            kwargs.get(
                "hosts",
                getattr(settings, "LISTENER_CONNECTION", {}).get(
                    "hosts", {"local": "127.0.0.1"}
                ),
            ).values()
        )

        self.port = kwargs.get(
            "port", getattr(settings, "LISTENER_CONNECTION", {}).get("port", 53892)
        )
        self.protocol = kwargs.get(
            "protocol",
            getattr(settings, "LISTENER_CONNECTION", {}).get("protocol", "tcp"),
        )

        self.max_concurrent_commands = kwargs.get(
            "max_concurrent_commands",
            getattr(settings, "LISTENER_CONNECTION", {}).get(
                "max_concurrent_commands", 10
            ),
        )

        # When zmq_socket kwarg is not None, use this one instead of creating
        # a new one.
        self.zmq_socket = kwargs.get("zmq_socket")

        self._listener_protocol: Optional[ListenerProtocol] = None

        # Running coordination.
        self._should_stop: Optional[asyncio.Event] = None
        self._runner_future: Optional[asyncio.Future] = None
        self._channels_listener: Optional[asyncio.Future] = None
        self._communicating_future: Optional[asyncio.Future] = None
        self._heartbeat_future: Optional[asyncio.Future] = None

    @property
    def should_stop(self):
        """Return should stop event.

        Use lazy instantiation to avoid creation event loop on init.
        """
        if self._should_stop is None:
            self._should_stop = asyncio.Event()
        return self._should_stop

    @property
    def hosts(self):
        """Get hosts to bind on."""
        return self._hosts

    @hosts.setter
    def hosts(self, hosts: dict):
        """Set the hosts to bind on."""
        self._hosts = list(hosts.values())

    @property
    def listener_protocol(self) -> ListenerProtocol:
        """Return the listener protocol object.

        Used to lazy create object when property is accessed.
        """
        if self._listener_protocol is None:
            self._listener_protocol = ListenerProtocol(
                self.hosts,
                self.port,
                self.protocol,
                self.zmq_socket,
                self.max_concurrent_commands,
            )
        return self._listener_protocol

    async def channels_listener(self):
        """Listen for terminate command and forward it to worker."""
        # Use different channel name when running tests.
        if is_testing():
            redis_prefix = getattr(settings, "FLOW_MANAGER", {}).get("REDIS_PREFIX", "")
            LISTENER_CONTROL_CHANNEL = "{}.listener".format(redis_prefix)  # noqa: F811

        logger.debug("Listener for terminate on channel %s", LISTENER_CONTROL_CHANNEL)
        channel_layer = get_channel_layer()
        while True:
            message = await channel_layer.receive(LISTENER_CONTROL_CHANNEL)
            message_type, peer_identity = message.get("type"), message.get("identity")
            if message_type == "terminate" and peer_identity is not None:
                logger.debug("Terminating peer '%s'.", peer_identity)
                await self.listener_protocol._message_processor.terminate(peer_identity)
            else:
                logger.error(__("Received unknown channels message '{}'.", message))

    async def __aenter__(self):
        """On entering a context, start the listener thread."""
        loop = asyncio.get_event_loop()
        loop.set_exception_handler(handle_exceptions)
        self._should_stop = None
        self._runner_future = asyncio.ensure_future(self.run())
        self._channels_listener = asyncio.ensure_future(self.channels_listener())
        self._heartbeat_future = asyncio.ensure_future(
            self.listener_protocol.heartbeat_task()
        )
        return self

    async def __aexit__(self, typ, value, trace):
        """On exiting a context, kill the listener and wait for it.

        .. note::

            Exceptions are all propagated.
        """
        logger.debug("Listener exiting context.")
        assert self._runner_future is not None
        assert self._channels_listener is not None
        assert self._heartbeat_future is not None
        self._channels_listener.cancel()
        self._heartbeat_future.cancel()
        self.terminate()
        logger.debug("Awaiting runner future.")
        await asyncio.gather(self._runner_future)
        self._listener_protocol = None
        logger.debug("Listener exited context.")

    def terminate(self):
        """Stop the standalone manager."""
        logger.info(__("Terminating Resolwe listener."))
        self.should_stop.set()

    async def run(self):
        """Run the main listener run loop.

        Doesn't return until :meth:`terminate` is called.
        """
        logger.info(
            __(
                f"Starting Resolwe listener on  '{self.protocol}://{self.hosts}:{self.port}'."
            )
        )
        communicator_future = asyncio.ensure_future(
            self.listener_protocol.communicate()
        )
        await self.should_stop.wait()
        self.listener_protocol.stop_communicate()
        await communicator_future
        logger.info(
            __(
                f"Stoping Resolwe listener on  '{self.protocol}://{self.hosts}:{self.port}'."
            )
        )



"""Protocol constants used by the manager and executors."""


class WorkerProtocol:
    """Constants used by the manager workers."""

    COMMAND = "command"
    DATA_ID = "data_id"

    COMMUNICATE = "communicate"
    COMMUNICATE_EXTRA = "kwargs"

    FINISH = "finish_data"
    FINISH_COMMUNICATE_EXTRA = "communicate_kwargs"

    ABORT = "abort_data"


class ExecutorProtocol:
    """Constants used by the executor<->listener protocol."""

    DOWNLOAD_STARTED = "download_started"
    DOWNLOAD_FINISHED = "download_finished"
    DOWNLOAD_ABORTED = "download_aborted"
    DOWNLOAD_IN_PROGRESS = "download_in_progress"
    DOWNLOAD_STARTED_LOCK = "download_started_lock"
    STORAGE_LOCATION_ID = "storage_location_id"
    STORAGE_ACCESS_LOG_ID = "storage_access_log_id"

    MISSING_DATA_LOCATIONS = "missing_data_locations"
    STORAGE_DATA_LOCATIONS = "storage_data_locations"

    GET_FILES_TO_DOWNLOAD = "get_files_to_download"
    REFERENCED_FILES = "referenced_files"

    RESULT = "result"
    RESULT_ERROR = "ER"


class ExecutorFiles:
    """Various files used by the executor."""

    # Base runtime subdirectory where settings will be stored.
    SETTINGS_SUBDIR = "settings"
    # Base runtime subdirectory where socket files will be.
    SOCKETS_SUBDIR = "sockets"

    FILE_LIST_KEY = "serialized_files"

    EXECUTOR_SETTINGS = "settings.json"
    DJANGO_SETTINGS = "django_settings.json"
    DATA = "data.json"
    LOCATION_SUBPATH = "storage_location.json"
    DATA_META = "data_meta.json"
    PROCESS = "process.json"
    PROCESS_META = "process_meta.json"

    PROCESS_SCRIPT = "process_script.sh"

    SECRETS_DIR = "secrets"

    STARTUP_PROCESSING_SCRIPT = "startup_processing_container.py"
    STARTUP_COMMUNICATION_SCRIPT = "startup_communication_container.py"
    SOCKET_UTILS = "socket_utils.py"
    CONSTANTS = "constants.py"



""".. Ignore pydocstyle D400.

==============================================================
Test if required file outputs of completed Data objects exists
==============================================================

Command to run on local machine::

    ./manage.py test_data_output

"""

import logging
from collections import defaultdict
from pathlib import Path
from typing import Set

from django.core.management.base import BaseCommand

from resolwe.flow.models import Data
from resolwe.flow.utils import iterate_fields

logger = logging.getLogger(__name__)

EXCLUDED_STORAGE_LOCATIONS = ["backup"]
LOG_PROGRESS = 5  # Log progress every LOG_PROGRESS%.


class Command(BaseCommand):
    """Test if file outputs of completed Data objects are referenced.

    We only check that the outputs are referenced in the database. The actual existence
    of the files has to be checked separately (see verify_data method on StoraLocation).
    """

    help = "Test if file outputs of completed Data objects exists."

    def get_files_dirs(
        self, output: dict, output_schema: dict
    ) -> tuple[set[str], set[str]]:
        """Get a set of files and dirs for a given output."""

        def normalize(path: str) -> str:
            """Normalize the path.

            For instance './file.txt' is translated into 'file.txt'.
            """
            return str(Path(path))

        paths_to_check: Set[str] = set()
        references: Set[str] = set()
        for field_schema, fields in iterate_fields(output, output_schema):
            name = field_schema["name"]
            value = fields[name]
            if "type" in field_schema:
                if field_schema["type"].startswith("basic:file:"):
                    paths_to_check.add(normalize(value["file"]))
                    references.update(normalize(ref) for ref in value.get("refs", []))

                elif field_schema["type"].startswith("list:basic:file:"):
                    for obj in value:
                        paths_to_check.add(normalize(obj["file"]))
                        references.update(normalize(ref) for ref in obj.get("refs", []))

                if field_schema["type"].startswith("basic:dir:"):
                    # Make sure '/' is appended if the path represents a directory.
                    to_add = normalize(value["dir"])
                    if not to_add.endswith("/"):
                        to_add += "/"
                    paths_to_check.add(to_add)
                    references.update(normalize(ref) for ref in value.get("refs", []))

                elif field_schema["type"].startswith("list:basic:dir:"):
                    for obj in value:
                        # Make sure '/' is appended if the path represents a directory.
                        to_add = normalize(obj["dir"])
                        if not to_add.endswith("/"):
                            to_add += "/"
                        paths_to_check.add(to_add)
                        references.update(
                            normalize(ref) for ref in value.get("refs", [])
                        )
        return paths_to_check, references

    def check_output(self, data: Data) -> bool:
        """Check if all files in data output are referenced."""
        logger.debug("Checking data with id %s.", data.id)
        assert data.status == Data.STATUS_DONE, (
            f"Data with id {data.id} has invalid status"
            "{data.status}, expected {Data.STATUS_DONE}."
        )
        assert data.location is not None, f"Data with id {data.id} has no location."
        paths, references = self.get_files_dirs(data.output, data.process.output_schema)

        # Check that the output files are referenced on all storage locations except backup.
        for storage_location in data.location.storage_locations.exclude(
            connector_name__in=EXCLUDED_STORAGE_LOCATIONS
        ):
            for path in paths:
                logger.debug(f"Checking {path}, location {storage_location}.")
                if not storage_location.files.filter(path=path).exists():
                    logger.error(
                        f"File {path} is missing on Data objects with id "
                        f"{data.pk}, storage location {storage_location}."
                    )
                    return False
            for reference in references:
                logger.debug(
                    f"Checking reference {reference}, location {storage_location}."
                )
                # Check that the reference exists, check for both file and directory.
                if (
                    not storage_location.files.filter(path=reference).exists()
                    and not storage_location.files.filter(path=reference + "/").exists()
                ):
                    logger.error(
                        f"Reference {reference} is missing on Data objects with id "
                        f"{data.pk}, storage location {storage_location}."
                    )
                    return False
        return True

    def add_arguments(self, parser):
        """Command arguments."""
        parser.add_argument(
            "-d",
            "--data",
            dest="data_ids",
            nargs="+",
            type=int,
            help="IDs of the data objects to check. Defaults to all completed data objects.",
            default=[],
        )

    def handle(self, *args, **kwargs):
        """Run test."""
        filters = {}
        if data_ids := kwargs["data_ids"]:
            filters = {"pk__in": data_ids}

        processed_location_ids: dict[int, list[int]] = defaultdict(list)
        failed_location_ids: set[int] = set()

        queryset = Data.objects.filter(status=Data.STATUS_DONE, **filters).exclude(
            location__isnull=True
        )

        to_process = queryset.count()
        previous_progress = 0.0
        progress = 0.0

        logger.info("Checking %d data objects.", to_process)
        for number, data in enumerate(queryset, start=1):
            logger.debug("Processing data %s/%s.", number, to_process)

            if data.location_id in processed_location_ids:
                logger.debug(
                    "Location of data with id %s already processed by data objects %s. Skipping.",
                    data.location_id,
                    processed_location_ids[data.location_id],
                )
            elif not self.check_output(data):
                failed_location_ids.add(data.location_id)
            processed_location_ids[data.location_id].append(data.id)

            # Log progress every LOG_PROGRESS%.
            progress = (number / to_process) * 100
            if progress - previous_progress >= LOG_PROGRESS:
                logger.info("Progress: %.2f%%", progress)
                previous_progress = progress

        if failed_location_ids:
            logger.error(
                "There are data objects with location IDs %s that have missing files. "
                "See error messages above for more information.",
                failed_location_ids,
            )
        else:
            logger.info("All files on outputs are present.")



"""Resolwe data serializer."""

from django.db import transaction
from rest_framework import serializers

from resolwe.flow.models import Collection, Data, DescriptorSchema, Entity, Process
from resolwe.permissions.models import Permission
from resolwe.rest.fields import ProjectableJSONField

from .base import ResolweBaseSerializer
from .collection import CollectionSerializer
from .descriptor import DescriptorSchemaSerializer
from .entity import BaseEntitySerializer
from .fields import DictRelatedField
from .process import ProcessSerializer


class DataSerializer(ResolweBaseSerializer):
    """Serializer for Data objects."""

    input = ProjectableJSONField(required=False)
    output = ProjectableJSONField(required=False)
    descriptor = ProjectableJSONField(required=False)
    process_resources = ProjectableJSONField(required=False)
    process = DictRelatedField(
        queryset=Process.objects.all(), serializer=ProcessSerializer
    )
    descriptor_schema = DictRelatedField(
        queryset=DescriptorSchema.objects.all(),
        serializer=DescriptorSchemaSerializer,
        allow_null=True,
        required=False,
    )
    collection = DictRelatedField(
        queryset=Collection.objects.all(),
        serializer=CollectionSerializer,
        allow_null=True,
        required=False,
        write_permission=Permission.EDIT,
    )
    entity = DictRelatedField(
        queryset=Entity.objects.all(),
        serializer=BaseEntitySerializer,
        allow_null=True,
        required=False,
        write_permission=Permission.EDIT,
    )

    class Meta:
        """DataSerializer Meta options."""

        model = Data
        read_only_fields = (
            "checksum",
            "created",
            "descriptor_dirty",
            "duplicated",
            "finished",
            "id",
            "modified",
            "process_cores",
            "process_error",
            "process_info",
            "process_memory",
            "process_progress",
            "process_rc",
            "process_warning",
            "output",
            "scheduled",
            "size",
            "started",
            "status",
        )
        update_protected_fields = (
            "contributor",
            "input",
            "process",
        )
        fields = (
            read_only_fields
            + update_protected_fields
            + (
                "collection",
                "descriptor",
                "descriptor_schema",
                "entity",
                "name",
                "slug",
                "tags",
                "process_resources",
            )
        )

    def validate_process(self, process):
        """Check that process is active."""
        if not process.is_active:
            raise serializers.ValidationError(
                "Process {} is not active.".format(process)
            )
        return process

    def validate(self, validated_data):
        """Validate collection change."""
        entity = validated_data.get("entity") or getattr(self.instance, "entity", None)
        collection = validated_data.get("collection") or getattr(
            self.instance, "collection", None
        )
        Data.validate_change_containers(self.instance, entity, collection)
        return super().validate(validated_data)

    @transaction.atomic
    def update(self, instance, validated_data):
        """Update the data object."""
        container_change = False
        entity = instance.entity
        collection = instance.collection
        if "collection" in validated_data:
            container_change = True
            collection = validated_data.pop("collection", None)
        if "entity" in validated_data:
            container_change = True
            entity = validated_data.pop("entity", None)
        instance = super().update(instance, validated_data)
        if container_change:
            instance.move_to_containers(entity, collection)
        return instance



"""Report the listener metrics to the monitoring system."""

import logging
import socket
from time import time
from typing import Any

from django.conf import settings
from opentelemetry import metrics
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk.resources import Resource

from resolwe.flow.executors.socket_utils import (
    Message,
    MessageProcessingCallback,
    MessageProcessingEventType,
    PeerIdentity,
)

logger = logging.getLogger(__name__)

# How often to export metrics (in milliseconds).
METRICS_EXPORT_INTERVAL = 30000
# TODO: Remove this when metrics are properly implemented.
METRICS_INSECURE_EXPORT = True


class MetricsEventReporter(MessageProcessingCallback):
    """Class handling the reporting of the metrics events."""

    # The prefix used in all metric names. The name is generated as
    # f"{self.metrics_prefix}_{metric_name}", so avoid adding '_' at the end.
    metrics_prefix = "resolwe_listener"

    def __init__(self):
        """Initialize the metrics event reporter."""
        if not (endpoint := getattr(settings, "FLOW_METRICS_ENDPOINT", None)):
            logger.info("Metrics endpoint is not set, reporting disabled.")
            self._enabled = False
            logger.warning(
                "Listener metrics endpoint is not set, diabling metrics reporting."
            )
        else:
            logger.info("Reporting metrics to endpoint %s.", endpoint)
            self._hostname = socket.gethostname()
            self._enabled = True
            self._init_metrics(endpoint)

    def _init_metrics(self, metric_endpoint):
        # Initialize the counters for the metrics.
        self._processing_messages = 0
        self._queued_messages = 0

        # Create the dictionaries with the appropriate timestamps.
        self._processing_started: dict[bytes, float] = {}
        self._message_received: dict[bytes, float] = {}

        # Initialize the opentelemetry metrics.
        metric_exporter = OTLPMetricExporter(
            endpoint=metric_endpoint,
            insecure=METRICS_INSECURE_EXPORT,
            max_export_batch_size=getattr(settings, "FLOW_METRICS_EXPORT_SIZE", 1000),
        )
        metric_reader = PeriodicExportingMetricReader(
            metric_exporter, METRICS_EXPORT_INTERVAL
        )
        provider = MeterProvider(
            metric_readers=[metric_reader],
            resource=Resource.create({"service.name": "listener"}),
        )

        # Sets the global default meter provider
        metrics.set_meter_provider(provider)

        # Creates a meter from the global meter provider
        meter = metrics.get_meter("listener")

        # Create the metrics.
        self._messages_queued = meter.create_up_down_counter(
            name=f"{self.metrics_prefix}_messages_queued",
            description="Number of unprocessed messages in the queue.",
        )
        self._messages_processing = meter.create_up_down_counter(
            name=f"{self.metrics_prefix}_messages_processing",
            description="Number of currently processing messages.",
        )
        self._preparation_time_histogram = meter.create_histogram(
            name=f"{self.metrics_prefix}_preparation_duration",
            description="The duration of init phase.",
            unit="seconds",
        )
        self._response_time_histogram = meter.create_histogram(
            name=f"{self.metrics_prefix}_response_duration",
            unit="miliseconds",
            description="Response time for messages.",
        )
        self._processing_time_histogram = meter.create_histogram(
            name=f"{self.metrics_prefix}_processing_duration",
            unit="miliseconds",
            description="Processing time for messages.",
        )
        self._network_time_histogram = meter.create_histogram(
            name=f"{self.metrics_prefix}_network_duration",
            unit="miliseconds",
            description="Time messages spent in the transfer.",
        )

    def _counter_changed(
        self, counter: metrics.UpDownCounter, value: int, attributes: dict
    ):
        """Change the value of the given counter."""
        counter.add(value, attributes)

    def _attributes_from_message(self, message: Message) -> dict[str, Any]:
        """Create the attributes from the given message."""
        return {"command": message.type_data}

    def event(
        self,
        event_type: MessageProcessingEventType,
        message: Message,
        peer_identity: PeerIdentity,
        **kwargs,
    ):
        """Process the event from the message processing pipeline."""
        # Do not proceed if metric reporting is disabled.
        if not self._enabled:
            return

        if message.client_id is None:
            logger.error("Message client_id is not defined.")
            return

        attributes = self._attributes_from_message(message)
        attributes["pod"] = self._hostname

        match event_type:
            case MessageProcessingEventType.MESSAGE_RECEIVED:
                self._queued_messages += 1
                self._counter_changed(self._messages_queued, 1, attributes)
                self._message_received[message.client_id] = time()
                value = (
                    self._message_received[message.client_id] - message.sent_timestamp
                )
                self._network_time_histogram.record(value * 1000, attributes)

            case MessageProcessingEventType.MESSAGE_PROCESSING_STARTED:
                self._processing_messages += 1
                self._counter_changed(self._messages_processing, 1, attributes)
                self._processing_started[message.client_id] = time()

            case MessageProcessingEventType.PREPARATION_FINISHED:
                # This type expects the 'started' unix timestamp in kwargs.
                if "started" not in kwargs:
                    logger.error(
                        "Extra data 'started' expected for preparation finished event."
                    )
                    return
                preparation_time = time() - kwargs["started"]
                self._preparation_time_histogram.record(preparation_time, attributes)

            case MessageProcessingEventType.MESSAGE_PROCESSING_FINISHED:
                self._processing_messages -= 1
                self._queued_messages -= 1
                self._counter_changed(self._messages_processing, -1, attributes)
                self._counter_changed(self._messages_queued, -1, attributes)

                if message.client_id not in self._processing_started:
                    logger.error(
                        "Message from client '%s' not found in the processing messages.",
                        message.client_id,
                    )
                    return
                if message.client_id not in self._message_received:
                    logger.error(
                        "Message from client '%s' not found in the received messages.",
                        message.client_id,
                    )
                    return

                processing_started = self._processing_started.pop(message.client_id)
                message_received = self._message_received.pop(message.client_id)
                processing_finished = time()
                response_time = processing_finished - message_received
                processing_time = processing_finished - processing_started
                self._response_time_histogram.record(response_time * 1000, attributes)
                self._processing_time_histogram.record(
                    processing_time * 1000, attributes
                )


metrics_reporter = MetricsEventReporter()



# Generated by Django 3.2.19 on 2023-06-05 07:22

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models

import resolwe.flow.models.fields
from resolwe.flow.models.fields import VersionField


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("flow", "0008_annotations"),
    ]

    operations = [
        migrations.AddField(
            model_name="annotationpreset",
            name="contributor",
            field=models.ForeignKey(
                default=None,
                on_delete=django.db.models.deletion.PROTECT,
                to="auth.user",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="annotationpreset",
            name="created",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AddField(
            model_name="annotationpreset",
            name="modified",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AddField(
            model_name="annotationpreset",
            name="slug",
            field=resolwe.flow.models.fields.ResolweSlugField(
                max_length=100, populate_from="name", unique_with=("version",)
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="annotationpreset",
            name="version",
            field=VersionField(default="0.0.0"),
        ),
        migrations.AlterField(
            model_name="annotationpreset",
            name="name",
            field=models.CharField(max_length=100),
        ),
    ]



# Generated by Django 3.2.16 on 2023-03-01 08:21

import django.db.models.deletion
from django.db import migrations, models

import resolwe.observers.models


class Migration(migrations.Migration):
    dependencies = [
        ("permissions", "0004_create_anonymous_user"),
        ("observers", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="BackgroundTask",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("started", models.DateTimeField(blank=True, db_index=True, null=True)),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("WT", "Waiting"),
                            ("PR", "Processing"),
                            ("OK", "Done"),
                            ("ER", "Error"),
                        ],
                        default="WT",
                        max_length=2,
                    ),
                ),
                (
                    "finished",
                    models.DateTimeField(blank=True, db_index=True, null=True),
                ),
                ("description", models.CharField(max_length=256)),
                (
                    "output",
                    models.JSONField(
                        default=resolwe.observers.models._default_output_value
                    ),
                ),
                (
                    "permission_group",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="backgroundtask",
                        to="permissions.permissiongroup",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



""".. Ignore pydocstyle D400.

==================
Permissions Filter
==================

"""

from django.db.models import QuerySet
from rest_framework.filters import BaseFilterBackend
from rest_framework.request import Request

from resolwe.permissions.models import PermissionQuerySet
from resolwe.permissions.utils import model_has_permissions


class ResolwePermissionsFilter(BaseFilterBackend):
    """Permissions filter."""

    def filter_queryset(self, request: Request, queryset: QuerySet, view) -> QuerySet:
        """Filter permissions queryset.

        When model has no permission group defined return the entire queryset.
        """
        if model_has_permissions(queryset.model):
            assert isinstance(queryset, PermissionQuerySet)
            queryset = queryset.filter_for_user(request.user)

        return queryset



"""Resolwe views utils."""

from rest_framework import exceptions, permissions

from resolwe.flow.models import Collection
from resolwe.permissions.models import Permission


def get_collection_for_user(collection_id, user):
    """Check that collection exists and user has `edit` permission."""
    collection_query = Collection.objects.filter(pk=collection_id)
    if not collection_query.exists():
        raise exceptions.ValidationError("Collection id does not exist")

    collection = collection_query.first()
    if not user.has_perm(Permission.EDIT, obj=collection):
        if user.is_authenticated:
            raise exceptions.PermissionDenied()
        else:
            raise exceptions.NotFound()

    return collection


class IsStaffuser(permissions.BasePermission):
    """Allow access only to staff users."""

    message = "Only staff users are allowed."

    def has_permission(self, request, view):
        """Return true when request is allowed."""
        return bool(
            request.user and request.user.is_authenticated and request.user.is_staff
        )


class IsStaffOrTargetUser(permissions.BasePermission):
    """Permission class for user endpoint."""

    def has_permission(self, request, view):
        """Check if user has permission."""
        return True

    def has_object_permission(self, request, view, obj):
        """Check if user has object permission."""
        return request.user.is_staff or obj == request.user


class IsSuperuserOrReadOnly(permissions.BasePermission):
    """Superuser has permissions, otherwise only safe methods are allowed."""

    def has_permission(self, request, view):
        """Check if user has permission."""
        return request.method in permissions.SAFE_METHODS or (
            request.user and request.user.is_superuser
        )


class IsStaffOrReadOnly(permissions.BasePermission):
    """Staff user has permissions, otherwise only safe methods are allowed."""

    def has_permission(self, request, view):
        """Check if user has permission."""
        return request.method in permissions.SAFE_METHODS or (
            request.user and request.user.is_staff
        )



"""Resolwe entity model."""

from itertools import batched
from typing import Any, List, Optional, Union

from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.indexes import GinIndex
from django.core.exceptions import ValidationError
from django.db import models, transaction

from resolwe.flow.models.annotations import AnnotationField, AnnotationValue
from resolwe.flow.models.base import BaseQuerySet
from resolwe.observers.consumers import BackgroundTaskType
from resolwe.observers.decorators import move_to_container
from resolwe.observers.models import BackgroundTask
from resolwe.observers.utils import start_background_task
from resolwe.permissions.models import PermissionObject

from .base import BaseModel
from .collection import BaseCollection, BaseCollectionQuerySet, Collection
from .utils import DirtyError, validate_schema


class EntityQuerySet(BaseCollectionQuerySet):
    """Query set for ``Entity`` objects."""

    def duplicate(self, request_user) -> BackgroundTask:
        """Duplicate (make a copy) ``Entity`` objects in the background."""
        task_data = {
            "entity_ids": list(self.values_list("pk", flat=True)),
            "inherit_collection": True,
        }
        return start_background_task(
            BackgroundTaskType.DUPLICATE_ENTITY,
            "Duplicate entity",
            task_data,
            request_user,
        )

    def delete_background(self, request_user):
        """Delete the ``Entity`` objects in the background."""
        task_data = {
            "object_ids": list(self.values_list("pk", flat=True)),
            "content_type_id": ContentType.objects.get_for_model(self.model).pk,
        }
        return start_background_task(
            BackgroundTaskType.DELETE, "Delete entities", task_data, request_user
        )

    def move_to_collection(self, destination_collection: Collection, request_user):
        """Move entities to destination collection."""
        task_data = {
            "target_id": destination_collection.pk,
            "data_ids": [],
            "entity_ids": list(self.values_list("pk", flat=True)),
            "request_user_id": request_user.pk,
        }
        return start_background_task(
            BackgroundTaskType.MOVE, "Move entities", task_data, request_user
        )

    def annotate_all(self, add_labels: bool = False):
        """Annotate with all metadata on the entities.

        The returned dataset has one entry for every annotation on the entity.
        It is annotated by "annotation_value" and the "annotation_field" in the
        form of f"{group_name}.{field_name}".

        :attr add_labels: when True and vocabulary is present, the annotation
            "annotation_label" contains the value of the label.
        """
        annotation_data = {
            "annotation_value": models.F("annotations__json__value"),
            "annotation_field": models.functions.Concat(
                models.F("annotations__field__group__name"),
                models.Value("."),
                models.F("annotations__field__name"),
            ),
        }
        if add_labels:
            annotation_data["annotation_label"] = models.functions.Coalesce(
                models.Func(
                    models.F("annotations__field__vocabulary"),
                    models.Func(
                        models.F("annotations__json"),
                        models.Value("value"),
                        function="jsonb_extract_path_text",
                    ),
                    function="jsonb_extract_path",
                    output_field=models.JSONField(),
                ),
                models.functions.Cast("annotations___value__value", models.JSONField()),
            )
        return self.annotate(**annotation_data)

    @classmethod
    def _prepare_annotation_data(
        cls,
        path: str,
        annotation_name: Optional[str] = None,
        value_to_label: bool = False,
        outerref_entity_pk_path: str = "pk",
    ):
        group_name, field_name = path.split(".", maxsplit=1)

        subquery = AnnotationValue.objects.filter(
            field__name=field_name,
            field__group__name=group_name,
            entity_id=models.OuterRef(outerref_entity_pk_path),
        )
        if value_to_label:
            subquery = subquery.annotate(
                final_value=models.functions.Coalesce(
                    models.Func(
                        models.F("field__vocabulary"),
                        models.Func(
                            models.F("_value"),
                            models.Value("value"),
                            function="jsonb_extract_path_text",
                        ),
                        function="jsonb_extract_path_text",
                        output_field=models.JSONField(),
                    ),
                    models.Func(
                        models.F("_value"),
                        models.Value("value"),
                        function="jsonb_extract_path_text",
                        output_field=models.JSONField(),
                    ),
                )
            )
        else:
            subquery = subquery.annotate(final_value=models.F("_value__value"))

        annotation_name = annotation_name or f"{group_name}_{field_name}"
        return {annotation_name: models.Subquery(subquery.values("final_value")[:1])}

    def annotate_path(
        self,
        paths: Union[str, List[str]],
        annotation_name: Optional[str] = None,
        value_to_label=False,
    ):
        """Add annotation to the Entity QuerySet.

        The annotation is the value of the field with the given name (in the
        given group).

        :attr paths: the path in the form 'group_name.field_name' or list of
            such paths.
        :attr field_name: the name of the annotation field.
        :attr annotation_name: the name under which annotation will be stored.
            When empty the name f"{group_name}_{field_name}" will be used. The
            annotation_name can only be used when path is a single string.
        :attr value_to_label: optionally annotate with label instead of the
            value. Only applicable when the vocabulary on the field is given.
        """
        if isinstance(paths, str):
            paths = [paths]
        queryset = self
        for path in paths:
            queryset = queryset.annotate(
                **self._prepare_annotation_data(path, annotation_name, value_to_label)
            )
        return queryset


class Entity(PermissionObject, BaseCollection):
    """Postgres model for storing entities."""

    class Meta(BaseCollection.Meta):
        """Entity Meta options."""

        permissions = (
            ("view", "Can view entity"),
            ("edit", "Can edit entity"),
            ("share", "Can share entity"),
            ("owner", "Is owner of the entity"),
        )

        indexes = [
            models.Index(name="idx_entity_name", fields=["name"]),
            GinIndex(
                name="idx_entity_name_trgm", fields=["name"], opclasses=["gin_trgm_ops"]
            ),
            models.Index(name="idx_entity_slug", fields=["slug"]),
            GinIndex(name="idx_entity_tags", fields=["tags"]),
            GinIndex(name="idx_entity_search", fields=["search"]),
        ]

    #: manager
    objects = EntityQuerySet.as_manager()

    #: collection to which entity belongs
    collection = models.ForeignKey(
        "Collection", blank=True, null=True, on_delete=models.CASCADE
    )

    #: entity type
    type = models.CharField(max_length=100, db_index=True, null=True, blank=True)

    #: duplication date and time
    duplicated = models.DateTimeField(blank=True, null=True)

    def get_annotation(self, path: str, default: Any = None) -> Any:
        """Get the annotation for the given path.

        :attr path: the path to the annotation in the format 'group.field'.
        :attr default: default value when annotation is not found.

        :return: value of the annotation or default if not found.
        """
        group_name, field_name = path.split(".", maxsplit=1)
        annotation: AnnotationValue = self.annotations.filter(
            field__group__name=group_name, field__name=field_name
        ).first()
        return annotation.value if annotation else default

    def set_annotation(self, path: str, value: Any, contributor=None):
        """Get the annotation for the given path.

        :attr path: the path to the annotation in the format 'group.field'.
        :attr value: the annotation value.
        """
        field_id = AnnotationField.id_from_path(path)
        contributor = contributor or self.contributor
        if field_id is not None:
            AnnotationValue.objects.create(
                field_id=field_id, entity=self, value=value, contributor=contributor
            )
        else:
            raise RuntimeError(f"No annotation field for path {path}.")

    def is_duplicate(self):
        """Return True if entity is a duplicate."""
        return bool(self.duplicated)

    def duplicate(self, request_user) -> BackgroundTask:
        """Duplicate (make a copy) object in the background."""
        task_data = {"entity_ids": [self.pk], "inherit_collection": True}
        return start_background_task(
            BackgroundTaskType.DUPLICATE_ENTITY,
            "Duplicate entity",
            task_data,
            request_user,
        )

    def delete_background(self):
        """Delete the object in the background."""
        task_data = {
            "object_ids": [self.pk],
            "content_type_id": ContentType.objects.get_for_model(self).pk,
        }
        return start_background_task(
            BackgroundTaskType.DELETE, "Delete entitiy", task_data, self.contributor
        )

    @transaction.atomic
    @move_to_container
    def move_to_collection(self, collection: Collection):
        """Move entity from the source to the destination collection.

        :args collection: the collection to move entity into.

        :raises ValidationError: when collection is set to None.
        """
        if self.collection == collection:
            return

        if collection is None:
            raise ValidationError(
                f"Entity {self}({self.pk}) can only be moved to another container."
            )

        if self.collection and (
            missing_annotations := self.collection.annotation_fields.exclude(
                pk__in=collection.annotation_fields.values_list("pk", flat=True)
            )
        ):
            collection.annotation_fields.add(*missing_annotations)

        self.collection = collection
        self.tags = collection.tags
        self.permission_group = collection.permission_group
        self.save(update_fields=["collection", "tags", "permission_group"])
        for data in self.data.all():
            data.move_to_collection(collection)

    def invalid_annotation_fields(self, annotation_fields=None):
        """Get the Queryset of invalid annotation fields.

        The invalid annotation field is a field that has annotatiton but it is
        not allowed in the collection this entity belongs to.

        :attr annotation_fields: the iterable containing annotations fields to
            be checked. When None is given the annotation fields belonging to
            the entity are checked.
        """
        annotation_fields = annotation_fields or self.collection.annotation_fields
        return AnnotationField.objects.filter(values=self.annotations).exclude(
            collection=self.collection
        )

    def copy_annotations(
        self, destination: "Entity", contributor
    ) -> List[AnnotationValue]:
        """Copy annotation from this entity to the destination.

        The objects are not saved to the database.
        """
        # Entity without collection can not have annotations.
        if self.collection is None:
            return []
        if destination.collection is None:
            raise ValidationError(
                f"The entity '{destination.slug}' must belong to the collection."
            )
        destination_annotations = []
        annotation_field_ids = set()
        # Create AnnotationValue objects from annotations and collect the set of
        # annatiton field ids for them. This set must also be available on the
        # destination entity, else exception is raised.
        for source_annotation in self.annotations.all():
            annotation_value = AnnotationValue(
                entity=destination,
                field_id=source_annotation.field_id,
                value=source_annotation.value,
                contributor=contributor or source_annotation.contributor,
            )
            destination_annotations.append(annotation_value)
            annotation_field_ids.add(source_annotation.field_id)
        return destination_annotations

    def validate_annotations(self):
        """Perform streamlined descriptor validation.

        :raises ValidationError: when annotations do not pass validation. All
            fields are validated and error messages aggregated into single
            exception.
        """
        invalid_annotation_fields = self.invalid_annotation_fields()
        if invalid_annotation_fields:
            msg = ",".join(field.label for field in invalid_annotation_fields)
            collection_id = getattr(self, "collection_id", None)
            raise ValidationError(
                f"The entity '{self.pk}' is annotated with fields '{msg}' "
                f"that are not allowed on the collection '{collection_id}'."
            )

        # Validate all the annotation values on this entity.
        validation_errors = []
        for annotation in self.annotations:
            try:
                annotation.validate()
            except ValidationError as validation_error:
                validation_errors.append(validation_error)
        if validation_errors:
            raise ValidationError(validation_errors)

    @transaction.atomic
    def update_annotations(
        self, annotations: dict[str, Any], contributor, update: bool = True
    ):
        """Update annotations with the given values.

        When annotation value is set no None the existing value is deleted.

        :attr annotations: the dictionary with annotation values. Keys are annotation
            paths.
        """
        # The annotations are processed in batches to avoid creating too complex
        # queries. The BATCH_SIZE determines the size of the batch.
        BATCH_SIZE = 1000
        field_map = {
            field_path: AnnotationField.field_from_path(field_path)
            for field_path in set(annotations.keys())
        }
        annotations = AnnotationValue.objects.add_annotations(
            [
                {
                    "field": field_map[field_path],
                    "entity": self,
                    "contributor": contributor,
                    "value": value,
                }
                for field_path, value in annotations.items()
            ]
        )
        # Old existing annotations must be deleted.
        if not update:
            created_entity_value_ids = {
                annotation.pk for annotation in annotations if not annotation.deleted
            }
            for pk_batch in batched(created_entity_value_ids, BATCH_SIZE):
                AnnotationValue.objects.filter(entity=self).exclude(
                    pk__in=created_entity_value_ids
                ).update(deleted=True)


class RelationType(models.Model):
    """Model for storing relation types."""

    #: relation type name
    name = models.CharField(max_length=100, unique=True)

    #: indicates if order of entities in relation is important or not
    ordered = models.BooleanField(default=False)

    def __str__(self):
        """Format the object representation."""
        return self.name


class Relation(BaseModel, PermissionObject):
    """Relations between entities.

    The Relation model defines the associations and dependencies between
    entities in a given collection:

    .. code-block:: json

        {
            "collection": "<collection_id>",
            "type": "comparison",
            "category": "case-control study",
            "entities": [
                {"enetity": "<entity1_id>", "label": "control"},
                {"enetity": "<entity2_id>", "label": "case"},
                {"enetity": "<entity3_id>", "label": "case"}
            ]
        }

    Relation ``type`` defines a specific set of associations among
    entities. It can be something like ``group``, ``comparison`` or
    ``series``. The relation ``type`` is an instance of
    :class:`.RelationType` and should be defined in any Django app that
    uses relations (e.g., as a fixture). Multiple relations of the same
    type are allowed on the collection.

    Relation ``category`` defines a specific use case. The relation
    category must be unique in a collection, so that users can
    distinguish between different relations. In the example above, we
    could add another ``comparison`` relation of ``category``, say
    ``Case-case study`` to compare ``<entity2>`` with ``<entity3>``.

    Relation is linked to :class:`resolwe.flow.models.Collection` to
    enable defining different relations structures in different
    collections. This also greatly speed up retrieving of relations, as
    they are envisioned to be mainly used on a collection level.

    ``unit`` defines units used in partitions where it is applicable,
    e.g. in relations of type ``series``.

    """

    UNIT_SECOND = "s"
    UNIT_MINUTE = "min"
    UNIT_HOUR = "hr"
    UNIT_DAY = "d"
    UNIT_WEEK = "wk"
    UNIT_CHOICES = (
        (UNIT_SECOND, "Second"),
        (UNIT_MINUTE, "Minute"),
        (UNIT_HOUR, "Hour"),
        (UNIT_DAY, "Day"),
        (UNIT_WEEK, "Week"),
    )

    #: type of the relation
    type = models.ForeignKey("RelationType", on_delete=models.PROTECT)

    #: collection to which relation belongs
    collection = models.ForeignKey("Collection", on_delete=models.CASCADE)

    #: partitions of entities in the relation
    entities = models.ManyToManyField(Entity, through="RelationPartition")

    #: category of the relation
    category = models.CharField(max_length=100)

    #: unit used in the partitions' positions (where applicable, e.g. for serieses)
    unit = models.CharField(max_length=3, choices=UNIT_CHOICES, null=True, blank=True)

    #: relation descriptor schema
    descriptor_schema = models.ForeignKey(
        "flow.DescriptorSchema", blank=True, null=True, on_delete=models.PROTECT
    )

    #: relation descriptor
    descriptor = models.JSONField(default=dict)

    #: indicate whether `descriptor` doesn't match `descriptor_schema` (is dirty)
    descriptor_dirty = models.BooleanField(default=False)

    #: custom manager with permission filtering methods
    objects = BaseQuerySet.as_manager()

    class Meta(BaseModel.Meta):
        """Relation Meta options."""

        unique_together = ("collection", "category")

        permissions = (
            ("view", "Can view relation"),
            ("edit", "Can edit relation"),
            ("share", "Can share relation"),
            ("owner", "Is owner of the relation"),
        )

    def save(self, *args, **kwargs):
        """Perform descriptor validation and save object."""
        if self.descriptor_schema:
            try:
                validate_schema(self.descriptor, self.descriptor_schema.schema)
                self.descriptor_dirty = False
            except DirtyError:
                self.descriptor_dirty = True
        elif self.descriptor and self.descriptor != {}:
            raise ValueError(
                "`descriptor_schema` must be defined if `descriptor` is given"
            )
        super().save()


class RelationPartition(models.Model):
    """Intermediate model to define Relations between Entities."""

    #: entity in the relation
    entity = models.ForeignKey(Entity, on_delete=models.CASCADE)

    #: relation to which entity belongs
    relation = models.ForeignKey(Relation, on_delete=models.CASCADE)

    #: number used for ordering entities in the relation
    position = models.PositiveSmallIntegerField(null=True, blank=True, db_index=True)

    #: Human-readable label of the position
    label = models.CharField(max_length=30, null=True, blank=True, db_index=True)

    class Meta:
        """RelationPartition Meta options."""

        unique_together = ("entity", "relation")



# pylint: disable=missing-docstring,too-many-lines
from datetime import timedelta

from django.utils.timezone import now

from resolwe.flow.models import (
    Collection,
    CollectionHistory,
    Data,
    DataHistory,
    Process,
)
from resolwe.flow.models.history import CollectionChange, SizeChange
from resolwe.flow.models.history_manager import history_manager
from resolwe.test import TestCase


class HistoryModelTest(TestCase):
    """Check that the history of data and collection objects is kept."""

    def test_data_history_created(self):
        process = Process.objects.create(contributor=self.contributor)
        data = Data.objects.create(
            name="data", contributor=self.contributor, process=process, size=0
        )
        current_time = now()
        history = data.history.get()
        self.assertIsInstance(history, DataHistory)
        size_history = history.sizechange.get()
        name_history = history.datanamechange.get()
        slug_history = history.dataslugchange.get()
        # There should be one history object with values as on data object.
        collection_history = history.collectionchange.get()
        self.assertEqual(size_history.value, data.size)
        self.assertEqual(name_history.value, data.name)
        self.assertEqual(slug_history.value, data.slug)
        self.assertEqual(collection_history.value, data.collection)
        # The timestamps of the history objects should be close to the current time.
        self.assertLess(size_history.timestamp - current_time, timedelta(seconds=1))
        self.assertLess(slug_history.timestamp - current_time, timedelta(seconds=1))
        self.assertLess(name_history.timestamp - current_time, timedelta(seconds=1))
        self.assertLess(
            collection_history.timestamp - current_time, timedelta(seconds=0.1)
        )

    def test_processing_history(self):
        process = Process.objects.create(contributor=self.contributor)
        data = Data.objects.create(
            name="data", contributor=self.contributor, process=process, size=0
        )
        history = data.history.get()
        self.assertEqual(history.processinghistory.count(), 0)

        data.started = now() - timedelta(hours=1)
        data.finished = now()
        data.save()
        processing_history = history.processinghistory.get()
        self.assertEqual(processing_history.allocated_cpu, data.process_cores)
        self.assertEqual(processing_history.allocated_memory, data.process_memory)
        self.assertEqual(processing_history.node_cpu, data.process_cores)
        self.assertEqual(processing_history.node_memory, data.process_memory)
        self.assertEqual(processing_history.interval.lower, data.started)
        self.assertEqual(processing_history.interval.upper, data.finished)

        data.process_cores += 1
        data.process_memory += 1
        data.finished = now() + timedelta(hours=1)
        data.save()
        self.assertEqual(history.processinghistory.count(), 2)
        processing_history = history.processinghistory.last()
        self.assertEqual(processing_history.allocated_cpu, data.process_cores)
        self.assertEqual(processing_history.allocated_memory, data.process_memory)
        self.assertEqual(processing_history.node_cpu, data.process_cores)
        self.assertEqual(processing_history.node_memory, data.process_memory)
        self.assertEqual(processing_history.interval.lower, data.started)
        self.assertEqual(processing_history.interval.upper, data.finished)

    def test_collection_history_created(self):
        collection = Collection.objects.create(
            name="collection", contributor=self.contributor
        )
        current_time = now()
        history = collection.history.get()
        self.assertIsInstance(history, CollectionHistory)
        with self.assertRaises(SizeChange.DoesNotExist):
            SizeChange.objects.get()

        with self.assertRaises(CollectionChange.DoesNotExist):
            CollectionChange.objects.get()

        name_history = history.collectionnamechange.get()
        slug_history = history.collectionslugchange.get()
        # There should be one history object with values as on data object.
        self.assertEqual(name_history.value, collection.name)
        self.assertEqual(slug_history.value, collection.slug)
        # The timestamps of the history objects should be close to the current time.
        self.assertLess(slug_history.timestamp - current_time, timedelta(seconds=1))
        self.assertLess(name_history.timestamp - current_time, timedelta(seconds=1))

    def test_data_history_updated(self):
        process = Process.objects.create(contributor=self.contributor)
        data = Data.objects.create(
            name="data", contributor=self.contributor, process=process, size=0
        )

        data.name = "new name"
        data.save()
        history = data.history.get()
        # There should be two namechange objects.
        self.assertEqual(history.datanamechange.count(), 2)
        self.assertEqual(history.datanamechange.first().value, "data")
        self.assertEqual(history.datanamechange.last().value, data.name)
        self.assertEqual(history.dataslugchange.count(), 1)
        self.assertEqual(history.sizechange.count(), 1)
        self.assertEqual(history.collectionchange.count(), 1)

        data.slug = "new-slug"
        data.save()
        history = data.history.get()
        # There should be two slugchange objects.
        self.assertEqual(history.dataslugchange.count(), 2)
        self.assertEqual(history.dataslugchange.first().value, "data")
        self.assertEqual(history.dataslugchange.last().value, data.slug)
        self.assertEqual(history.datanamechange.count(), 2)
        self.assertEqual(history.sizechange.count(), 1)
        self.assertEqual(history.collectionchange.count(), 1)

        data.size = 10
        data.save()
        history = data.history.get()
        # There should be two sizechange objects.
        self.assertEqual(history.sizechange.count(), 2)
        self.assertEqual(history.sizechange.first().value, 0)
        self.assertEqual(history.sizechange.last().value, data.size)
        self.assertEqual(history.datanamechange.count(), 2)
        self.assertEqual(history.dataslugchange.count(), 2)
        self.assertEqual(history.sizechange.count(), 2)
        self.assertEqual(history.collectionchange.count(), 1)

        collection = Collection.objects.create(
            name="collection", contributor=self.contributor
        )
        data.collection = collection
        data.save()
        history = data.history.get()
        # There should be two collectionchange objects.
        self.assertEqual(history.collectionchange.count(), 2)
        self.assertIsNone(history.collectionchange.first().value)
        self.assertEqual(
            history.collectionchange.last().value, collection.history.get()
        )
        self.assertEqual(history.datanamechange.count(), 2)
        self.assertEqual(history.dataslugchange.count(), 2)
        self.assertEqual(history.sizechange.count(), 2)
        self.assertEqual(history.collectionchange.count(), 2)

    def test_collection_history_updated(self):
        collection = Collection.objects.create(
            name="collection", contributor=self.contributor
        )

        collection.name = "new name"
        collection.save()
        history = collection.history.get()
        # There should be two namechange objects.
        self.assertEqual(history.collectionnamechange.count(), 2)
        first_name_change = history.collectionnamechange.first()
        last_name_change = history.collectionnamechange.last()
        self.assertIsNotNone(first_name_change)
        self.assertIsNotNone(last_name_change)
        # Add asserts bellow to keep mypy happy.
        assert first_name_change is not None
        assert last_name_change is not None
        self.assertEqual(first_name_change.value, "collection")
        self.assertEqual(last_name_change.value, collection.name)
        self.assertEqual(history.collectionnamechange.count(), 2)
        self.assertEqual(history.collectionslugchange.count(), 1)

        collection.slug = "new-slug"
        collection.save()
        history = collection.history.get()
        # There should be two slugchange objects.
        first_slug_change = history.collectionslugchange.first()
        last_slug_change = history.collectionslugchange.last()
        self.assertIsNotNone(first_slug_change)
        self.assertIsNotNone(last_slug_change)
        # Add asserts bellow to keep mypy happy.
        assert first_slug_change is not None
        assert last_slug_change is not None
        self.assertEqual(history.collectionslugchange.count(), 2)
        self.assertEqual(first_slug_change.value, "collection")
        self.assertEqual(last_slug_change.value, collection.slug)
        self.assertEqual(history.collectionnamechange.count(), 2)
        self.assertEqual(history.collectionslugchange.count(), 2)

    def test_history_values(self):
        """Test getting values at certain timestamp."""
        collection = Collection.objects.create(
            name="collection", contributor=self.contributor
        )
        collection.name = "new name"
        collection.save()
        history = collection.history.get()
        self.assertEqual(history.collectionnamechange.count(), 2)
        first_change = history.collectionnamechange.first()
        second_change = history.collectionnamechange.last()
        self.assertIsNotNone(first_change)
        self.assertIsNotNone(second_change)
        # Add asserts bellow to keep mypy happy.
        assert first_change is not None
        assert second_change is not None
        second_change.timestamp = second_change.timestamp + timedelta(seconds=2)
        second_change.save(update_fields=["timestamp"])
        before_timestamp = first_change.timestamp - timedelta(seconds=1)
        after_timestamp = second_change.timestamp + timedelta(seconds=1)
        between_timestamp = first_change.timestamp + timedelta(seconds=1)
        # Reading value before the first change should raise RuntimeError.
        with self.assertRaises(RuntimeError):
            history_manager.valid_value(collection, "name", before_timestamp)
        value_between = history_manager.valid_value(
            collection, "name", between_timestamp
        )
        value_after = history_manager.valid_value(collection, "name", after_timestamp)

        self.assertEqual(value_between, "collection")
        self.assertEqual(value_after, "new name")

    def test_delete(self):
        """Test deletion is tracked."""
        collection = Collection.objects.create(
            name="collection", contributor=self.contributor
        )
        history = collection.history.get()
        self.assertIsNotNone(history.datum)
        self.assertIsNone(history.deleted)
        collection.delete()
        history.refresh_from_db()
        current_time = now()
        self.assertIsNone(history.datum)
        self.assertIsNotNone(history.deleted)
        assert history.deleted is not None  # keep mypy happy
        self.assertLess(current_time - history.deleted, timedelta(seconds=1))



"""Registry class for storage connectors."""

import logging
from typing import Dict, MutableMapping, Sequence, Type

from .baseconnector import DEFAULT_CONNECTOR_PRIORITY, BaseStorageConnector

logger = logging.getLogger(__name__)


class StorageConnectors(MutableMapping[str, BaseStorageConnector]):
    """A class holding configured instances of storage connectors.

    The connector classes are registered automatically via the hook
    __init_subclass__ in the BaseStorageConnector class when connector classes
    are imported.

    This class is implemented as singleton: constructing multiple instances
    of it will always return the same instance.
    """

    def __init__(self):
        """Initialize."""
        self._connectors: Dict[str, BaseStorageConnector] = dict()
        self._connector_classes: Dict[str, Type[BaseStorageConnector]] = dict()
        self._storage_connectors: Dict[str, Sequence[BaseStorageConnector]] = dict()

    def _get_connectors_settings(self) -> dict:
        """Return settings for storage connectors.

        This module can also be imported from inside executor special care
        should be taken from where settings are read.
        """
        key = "STORAGE_CONNECTORS"
        try:
            # Try to import settings from executor.
            from executors.global_settings import SETTINGS

            settings = SETTINGS.get(key, {})
        except ImportError:
            # Import settings from Django.
            from resolwe.storage.settings import (
                STORAGE_CONNECTORS as settings,  # type: ignore
            )

        return settings

    def _recreate_storage_mappings(self):
        """Recreate mapping between storages and connectors."""
        try:
            # Try to import settings from executor.
            from executors.global_settings import SETTINGS

            storages = SETTINGS.get("FLOW_STORAGE", dict())
        except ImportError:
            # Import settings from Django.
            from resolwe.storage import settings as storage_settings

            storages = storage_settings.FLOW_STORAGE

        self._storage_connectors = dict()
        for storage_name, storage_config in storages.items():
            self._storage_connectors[storage_name] = sorted(
                (
                    self[connector_name]
                    for connector_name in storage_config["connectors"]
                    if connector_name in self
                ),
                key=lambda connector: connector.config.get(
                    "priority", DEFAULT_CONNECTOR_PRIORITY
                ),
            )

    def add_storage_connector_class(self, connector: Type[BaseStorageConnector]):
        """Add storage connector class to the registry."""

        def _get_class_name(klass: Type[BaseStorageConnector]):
            return "{}.{}".format(klass.__module__, klass.__name__)

        assert issubclass(connector, BaseStorageConnector)
        full_class_name = _get_class_name(connector)
        logger.info("Adding connector {} to register".format(full_class_name))
        self._connector_classes[full_class_name] = connector
        self._add_connectors_from_settings(full_class_name, connector)

    def _add_connectors_from_settings(
        self, full_class_name: str, klass: Type[BaseStorageConnector]
    ):
        """Create instances for the connector of the given type.

        Iterate through connectors defined in the settings for the given
        class name, create their instances and add them to the registry.
        """
        connectors_settings = self._get_connectors_settings()
        for connector_name, connector_settings in connectors_settings.items():
            if connector_settings["connector"] == full_class_name:
                # Validate connector config before creating instance.
                for setting_name in klass.REQUIRED_SETTINGS:
                    if setting_name not in connector_settings["config"]:
                        raise Exception(
                            "Setting named {} must be set in connector {} configuration".format(
                                setting_name, connector_name
                            )
                        )
                instance = klass(connector_settings["config"], connector_name)
                self[connector_name] = instance
        self._recreate_storage_mappings()

    def for_storage(self, storage: str) -> Sequence[BaseStorageConnector]:
        """Get storage connectors for storage ordered by priority."""
        return self._storage_connectors[storage]

    def recreate_connectors(self):
        """Recreate connectors from settings."""
        for class_name, class_ in self._connector_classes.items():
            self._add_connectors_from_settings(class_name, class_)

    def __getitem__(self, item):
        """Get implementation."""
        return self._connectors.get(item)

    def __iter__(self):
        """Return iterator."""
        return iter(self._connectors)

    def __contains__(self, name):
        """Get if registry contains object with the given name."""
        return name in self._connectors

    def __len__(self) -> int:
        """Return the number of registered connectors."""
        return len(self._connectors)

    def __setitem__(self, name, connector):
        """Register or update new connector."""
        assert isinstance(connector, BaseStorageConnector)
        self._connectors[name] = connector

    def __delitem__(self, name):
        """Remove registered connector."""
        self._connectors.pop(name)


connectors = StorageConnectors()



"""Resolwe annotations model."""

import abc
import datetime
import re
from collections import defaultdict
from enum import Enum
from functools import reduce
from itertools import batched
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    List,
    Mapping,
    MutableSequence,
    Optional,
    Sequence,
    Type,
    TypedDict,
)

from django.core.exceptions import ValidationError
from django.db import models, transaction

from resolwe.flow.models.base import (
    MAX_SLUG_LENGTH,
    BaseManager,
    BaseModel,
    ResolweSlugField,
)
from resolwe.flow.models.fields import VersionField
from resolwe.permissions.models import (
    PermissionInterface,
    PermissionObject,
    PermissionQuerySet,
)

if TYPE_CHECKING:
    from resolwe.flow.models import Collection, Entity

from .base import VERSION_NUMBER_BITS, AuditModel

VALIDATOR_LENGTH = 128
NAME_LENGTH = 128
LABEL_LENGTH = 128
DESCRIPTION_LENGTH = 256


class AnnotationType(Enum):
    """Supported annotation types."""

    DATE = "DATE"
    DECIMAL = "DECIMAL"
    INTEGER = "INTEGER"
    STRING = "STRING"


class AnnotationValueValidator:
    """Class that validates the annotation value.

    It uses plugins (classes) to validate the annotation of the given type.
    """

    def __init__(self):
        """Initialize."""
        self._validators: Mapping[
            AnnotationType, MutableSequence["AnnotationFieldBaseValidator"]
        ] = defaultdict(list)

    def _add_validator(self, validator: "AnnotationFieldBaseValidator"):
        """Add the given validator."""
        for field_type in validator.annotation_field_types:
            self._validators[field_type].append(validator)

    def validate(
        self, annotation_value: "AnnotationValue", raise_exception: bool = False
    ) -> Sequence[ValidationError]:
        """Validate the given AnnotationValue object.

        All validation errors are grouped together.

        :raises ValidationError: if annotation fails and raise_exception is True. The
            actual errors are nested inside outer ValidationError.
        """
        errors = []
        validators = self._validators[annotation_value.field.annotation_type]
        for validator in validators:
            try:
                validator.validate(annotation_value.value, annotation_value.field)
            except ValidationError as error:
                errors.append(error)
        if errors and raise_exception:
            raise ValidationError(errors)
        return errors


annotation_value_validator = AnnotationValueValidator()


class AnnotationFieldBaseValidator(metaclass=abc.ABCMeta):
    """Base class for annotation class validation."""

    annotation_field_types: Iterable[AnnotationType]

    @classmethod
    def __init_subclass__(cls: Type["AnnotationFieldBaseValidator"], **kwargs):
        """Register instance of the class with the validator."""
        super().__init_subclass__(**kwargs)
        annotation_value_validator._add_validator(cls())

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate the value.

        :raises ValidatiorError: on validation failure.
        """
        raise NotImplementedError("Subclass should implement the 'validate' method.")


class AnnotationFieldStringValidator(AnnotationFieldBaseValidator):
    """Validates the string field."""

    annotation_field_types = (AnnotationType.STRING,)

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate value to be of type str."""
        if not isinstance(value, str):
            raise ValidationError(
                f"The value '{value}' is not of the expected type 'str'."
            )


class AnnotationFieldIntegerValidator(AnnotationFieldBaseValidator):
    """Validates the integer field."""

    annotation_field_types = (AnnotationType.INTEGER,)

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate value to be of type int."""
        if not isinstance(value, int):
            raise ValidationError(
                f"The value '{value}' is not of the expected type 'int'."
            )


class AnnotationFieldDecimalValidator(AnnotationFieldBaseValidator):
    """Validates the decimal field."""

    annotation_field_types = (AnnotationType.DECIMAL,)

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate value to be of type int or float."""
        if not isinstance(value, (int, float)):
            raise ValidationError(
                f"The value '{value}' is not of the expected types 'int' or 'float'."
            )


class AnnotationFieldDateValidator(AnnotationFieldBaseValidator):
    """Validates the date field."""

    annotation_field_types = (AnnotationType.DATE,)

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate value represents a date in the format YYYY-MM-DD.

        The month and day part can contain a single character.
        """
        if not isinstance(value, str):
            raise ValidationError(
                f"The value '{value}' is not of the expected type 'str'."
            )
        try:
            datetime.datetime.strptime(value, "%Y-%m-%d")
        except ValueError:
            raise ValidationError(
                f"Value {value} has incorrect format, use YYYY-MM-DD."
            )


class AnnotationFieldRegexValidator(AnnotationFieldBaseValidator):
    """Validates regex for the annotation value."""

    annotation_field_types = (
        AnnotationType.INTEGER,
        AnnotationType.DATE,
        AnnotationType.DECIMAL,
        AnnotationType.STRING,
    )

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate that the value matches regex defined on the field.

        The value is converted to string before matching.
        """
        # Validate the regex.
        if annotation_field.validator_regex is not None:
            regex = annotation_field.validator_regex
            value = str(value)
            if re.search(regex, value) is None:
                raise ValidationError(
                    f"The value '{value}' for the field '{annotation_field.pk}' does "
                    f"not match the regex '{regex}'."
                )


class AnnotationFieldVocabularyValidator(AnnotationFieldBaseValidator):
    """Validates that the value matches vocabulary."""

    annotation_field_types = (
        AnnotationType.INTEGER,
        AnnotationType.DATE,
        AnnotationType.DECIMAL,
        AnnotationType.STRING,
    )

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate that the value matches vocabulary on the field."""
        if (
            annotation_field.vocabulary is not None
            and value not in annotation_field.vocabulary
        ):
            raise ValidationError(
                f"The value '{value}' is not valid for the field {annotation_field}."
            )


class AbstractAnnotationGroup(models.Model):
    """Abstract base class for annotation groups."""

    #: the name of the annotation group
    name = models.CharField(max_length=NAME_LENGTH)

    #: the label of the annotation group
    label = models.CharField(max_length=LABEL_LENGTH)

    #: the sorting order among annotation groups
    sort_order = models.PositiveSmallIntegerField()

    class Meta:
        """Make this class abstract."""

        abstract = True


class AnnotationGroup(AbstractAnnotationGroup):
    """Group of annotation fields."""

    class Meta:
        """Set the default ordering."""

        ordering = ["sort_order"]

    def __str__(self) -> str:
        """Return user-friendly string representation."""
        return f"{self.name}"


class AnnotationFieldManager(
    BaseManager["AnnotationField", models.QuerySet["AnnotationField"]]
):
    """Annotation field manager.

    Return only the latest version of each field.
    """

    partition_fields = ["name", "group"]


class AbstractAnnotationField(models.Model):
    """Abstract base class for annotation fields."""

    partition_fields = ["name", "group_id"]
    version_field = "version"

    #: the name of the annotation fields
    name = models.CharField(max_length=NAME_LENGTH)

    #: user visible field name
    label = models.CharField(max_length=LABEL_LENGTH)

    #: user visible field description
    description = models.CharField(max_length=DESCRIPTION_LENGTH)

    #: the type of the annotation field
    type = models.CharField(max_length=16)

    #: the annotation group this field belongs to
    group = models.ForeignKey(
        AnnotationGroup, on_delete=models.CASCADE, related_name="fields"
    )

    #: the sorting order among annotation fields
    sort_order = models.PositiveSmallIntegerField()

    #: optional regular expression for validation
    validator_regex = models.CharField(
        max_length=VALIDATOR_LENGTH, null=True, blank=True
    )

    #: optional map of valid values to labels
    vocabulary = models.JSONField(null=True, blank=True)

    #: is this field required
    required = models.BooleanField(default=False)

    #: field version
    version = VersionField(number_bits=VERSION_NUMBER_BITS, default="0.0.0")

    class Meta:
        """Make this class abstract."""

        abstract = True

    def __init__(self, *args, **kwargs):
        """Store original vocabulary to private variable."""
        super().__init__(*args, **kwargs)
        self._original_vocabulary = self.vocabulary
        self._original_type = self.type

    @staticmethod
    def add_to_collection(source: "Collection", destination: "Collection"):
        """Add fields from the source to the destination collection."""
        destination.annotation_fields.add(*source.annotation_fields.all())

    @property
    def annotation_type(self) -> AnnotationType:
        """Get the field type as enum."""
        return AnnotationType(self.type.upper())

    @staticmethod
    def group_field_from_path(path: str) -> List[str]:
        """Return the group and field name from path."""
        match = re.fullmatch(
            r"(?P<group>[a-zA-Z][\w ]+)\.(?P<field>[a-zA-Z][\w ]+)", path, re.ASCII
        )
        if match is None:
            raise ValidationError(f"Invalid path '{path}'.")
        return [match["group"], match["field"]]

    @property
    def path(self) -> str:
        """Return the path of the field."""
        return f"{self.group.name}.{self.name}"

    @classmethod
    def id_from_path(cls, path: str) -> Optional[int]:
        """Get the field id from the field path."""
        group_name, field_name = cls.group_field_from_path(path)
        return (
            cls.objects.filter(group__name=group_name, name=field_name)
            .values_list("id", flat=True)
            .first()
        )

    @classmethod
    def field_from_path(cls, path: str) -> "AnnotationField":
        """Get the field id from the field path.

        :raises ValidationError: when field does not exist.
        """
        group_name, field_name = cls.group_field_from_path(path)
        field = cls.objects.filter(group__name=group_name, name=field_name).first()
        if not field:
            raise ValidationError(f"Field '{path}' does not exist.")
        return field

    def save(self, *args, **kwargs):
        """Recompute the labels for annotation values if vocabulary changes.

        :raises ValidationError: when vocabulary changes so that annotation values are no
            longer valid.
        """
        super().save(*args, **kwargs)
        if self.vocabulary != self._original_vocabulary:
            updated_fields = []
            for annotation_value_field in self.values.all():
                annotation_value_field.recompute_label()
                annotation_value_field.validate()
                updated_fields.append(annotation_value_field)
            AnnotationValue.all_objects.bulk_update(updated_fields, ["_value"])
            self._original_vocabulary = self.vocabulary
        if self.type != self._original_type:
            self.revalidate_values()
            self._original_type = self.type

    def revalidate_values(self):
        """Revalidate all annotation values.

        :raises ValidationError: when validation fails.
        """
        for annotation_value in self.values.iterator():
            annotation_value.validate()

    def label_by_value(self, label: str) -> str:
        """Get the value by label.

        When no value is found the label is returned.
        """
        if self.vocabulary is not None:
            for vocabulary_value, vocabulary_label in self.vocabulary.items():
                if label == vocabulary_label:
                    return vocabulary_value
        return label


class AnnotationField(AbstractAnnotationField):
    """Annotation field."""

    #: the latest version of the models
    objects = AnnotationFieldManager()

    #: all models
    all_objects = models.Manager()

    class Meta:
        """Set the constraints and the default ordering."""

        constraints = [
            # Accept only supported annotation types.
            models.constraints.CheckConstraint(
                check=models.Q(type__in=[e.value for e in AnnotationType]),
                name="annotation_type",
            ),
            models.constraints.UniqueConstraint(
                fields=["name", "group", "version"],
                name="uniquetogether_name_group_version",
            ),
        ]
        ordering = ["group__sort_order", "sort_order"]

    def __str__(self) -> str:
        """Return user-friendly string representation."""
        return f"{self.group.name}.{self.name}"


class AnnotationPreset(PermissionObject, BaseModel):
    """The named set of annotation fields.

    The presets have permissions.
    """

    #: the fields belonging to this preset
    fields = models.ManyToManyField(AnnotationField, related_name="presets")

    def __str__(self):
        """Return the string representation."""
        return self.name

    def can_set_permission(self) -> bool:
        """Return whether permissions can be set."""
        return True

    class Meta:
        """Override parent meta."""


class AddAnnotationDict(TypedDict):
    """The type of the dictionary for bulk adding annotations."""

    field: AnnotationField
    entity: "Entity"
    contributor: Any
    value: Any


class AnnotationValueManager(BaseManager["AnnotationValue", PermissionQuerySet]):
    """Annotation value manager.

    Return only the latest version of each field.
    """

    QuerySet = PermissionQuerySet
    enable_versioning = False

    def get_queryset(self):
        """Return only the values with delete marker not set."""
        return super().get_queryset().filter(deleted=False)

    def add_annotations(
        self, annotations: Iterable[AddAnnotationDict]
    ) -> list["AnnotationValue"]:
        """Bulk add annotations.

        We have to consider the possibilities:
        - the value is not a delete marker:
          - the existing value exists:
            - create a new entry and set deleted to true on the existing one if the value
              has actually changed.
          - the value does not exist yet or is deleted: create a new entry.
        - the value is a delete marker:
          - the existing value exists: set deleted to true.
        """

        def is_delete_marker(value):
            """Is the value a delete marker."""
            return value is None

        def combine_query(query: models.Q, data: dict) -> models.Q:
            """Combine the query with the data."""
            return query | models.Q(field=data["field"], entity=data["entity"])

        # Make method more general for use in classed derived from AnnotationValue.
        Model = self.model
        # The annotations are processed in batches to avoid creating too complex
        # queries. The BATCH_SIZE determines the size of the batch.
        BATCH_SIZE = 1000
        created = []
        for batch in batched(annotations, BATCH_SIZE):
            to_process = {(data["field"].pk, data["entity"].pk): data for data in batch}
            # Create a mapping to easy access the existing values.
            existing_values = {
                (value.field_id, value.entity_id): value
                for value in Model.objects.filter(
                    reduce(combine_query, batch, models.Q())
                )
            }
            # We have to:
            # - create new values
            # - create new values and delete existing ones if the value has changed
            # - delete values with None
            to_delete = []
            to_create = []
            for key, data in to_process.items():
                if existing_value := existing_values.get(key):
                    if is_delete_marker(data["value"]):
                        to_delete.append(existing_value.pk)
                    elif existing_value.value != data["value"]:
                        to_create.append(data)
                        to_delete.append(existing_value.pk)
                elif not is_delete_marker(data["value"]):
                    to_create.append(data)

            # Create new objects we must call the save method to set the delete marker.
            created_values = [Model(**data) for data in to_create]
            for value in created_values:
                value.save()

            created.extend(created_values)
            # Mark objects as deleted.
            if to_delete:
                Model.objects.filter(pk__in=to_delete).update(deleted=True)

        # Only return created objects.
        return created


def _slug_for_annotation_value(instance: "AnnotationValue") -> str:
    """Generate the slug for the annotation value."""
    return f"{instance.entity.slug}-{instance.field.group.name}-{instance.field.name}"


class AbstractAnnotationValue(BaseModel):
    """Abstract base class for annotation values."""

    partition_fields = ["field", "entity"]
    version_field = "created"

    #: URL slug
    slug = ResolweSlugField(
        populate_from=_slug_for_annotation_value,
        unique_with="version",
        max_length=MAX_SLUG_LENGTH,
    )

    #: the entity this field belongs to
    entity: "Entity" = models.ForeignKey(
        "Entity", related_name="annotations", on_delete=models.CASCADE
    )  # type: ignore

    #: the field this field belongs to
    field = models.ForeignKey(
        AnnotationField, related_name="values", on_delete=models.PROTECT
    )

    #: the delete marker
    deleted = models.BooleanField(default=False)

    #: value is stored under key 'value' in the json field to simplify lookups
    #: the value None is used as delete marker
    _value: Any = models.JSONField(default=dict, null=True)

    class Meta:
        """Make this class abstract."""

        abstract = True

    def __init__(self, *args, **kwargs):
        """Allow us to set the 'value' in the constructor.

        This is a little tricky, as value can be passed in args or kwargs. We only
        intercept the value when given as kwarg, since args are used by Django when
        constructing class from database value.
        """
        if "value" in kwargs:
            kwargs["_value"] = {"value": kwargs.pop("value")}
        super().__init__(*args, **kwargs)
        if "label" not in self._value:
            self.recompute_label()

    @property
    def delete_marker(self) -> bool:
        """Return True if this value represents a delete marker."""
        return self.deleted

    @property
    def value(self) -> str | int | float | datetime.date:
        """Get the actual value or None if object is delete marker."""
        return self._value["value"]

    @value.setter
    def value(self, value: str | int | float | datetime.date):
        """Set the value.

        The label is recomputed when the value is set.
        """
        self._value["value"] = value
        self.recompute_label()

    @property
    def _computed_label(self) -> Any:
        """Return computed label from value.

        :raises KeyError: when the value is not in the vocabulary.
        """
        if self.field.vocabulary is None:
            return self.value
        else:
            # Always return a value even if it is not in the vocabulary. The validation
            # step will take care of the wrong values.
            return self.field.vocabulary.get(self.value, self.value)

    @property
    def label(self) -> Any:
        """Return the cached label."""
        return self._value["label"]

    def validate(self):
        """Validate the given value.

        The validation is always done in full, all errors are gathered and sent
        in the response.

        :raises ValidationError: when the validation fails.
        """
        if not self.delete_marker:
            annotation_value_validator.validate(self, raise_exception=True)

    @transaction.atomic
    def save(self, *args, **kwargs):
        """Save the annotation value after validation has passed."""
        # Set deleted on other instances with the same field and entity.
        if self._state.adding:
            type(self).all_objects.filter(
                field=self.field, entity=self.entity, deleted=False
            ).update(deleted=True)

        if not self.delete_marker:
            annotation_value_validator.validate(self, raise_exception=True)
            # Make sure the label is always set.
            if "label" not in self._value:
                self.recompute_label()
        super().save(*args, **kwargs)

    def recompute_label(self):
        """Recompute label from value and set it to the model instance."""
        self._value["label"] = self._computed_label

    @staticmethod
    def from_path(entity_id: int, path: str) -> Optional["AnnotationValue"]:
        """Get the annotation value from the path."""
        field_id = AnnotationField.id_from_path(path)
        return AnnotationValue.objects.filter(
            entity_id=entity_id, field_id=field_id
        ).first()

    @classmethod
    def permission_proxy(cls) -> str:
        """Return the permission group path."""
        return "entity"

    def in_container(self) -> bool:
        """Return if object lies in a container."""
        return True


class AnnotationValue(AbstractAnnotationValue, PermissionInterface, AuditModel):
    """The value of the annotation."""

    #: the latest version of the models
    objects = AnnotationValueManager()  # type: ignore

    #: all models
    all_objects: models.Manager["AnnotationValue"] = PermissionQuerySet.as_manager()

    class Meta:
        """Set the unique constraints."""

        constraints = [
            models.constraints.UniqueConstraint(
                fields=["entity", "field", "created"],
                name="uniquetogether_entity_field_created",
            ),
        ]
        ordering = ["field__group__sort_order", "field__sort_order"]

    def __str__(self) -> str:
        """Return user-friendly string representation."""
        if self.delete_marker:
            return "Delete marker"
        else:
            return f"{self.label}"



"""Resolwe custom middleware class."""

import logging
import uuid

import resolwe.auditlog.auditmanager
from resolwe.permissions.models import get_anonymous_user

logger = logging.getLogger(__name__)


class ResolweAuditMiddleware:
    """Register custom audit manager.

    The manager looks for object request (read/write/create) by modifying base
    object and registering with the audit manager in appropriate moments.

    The manager is registered at the beginning of the request and storage into
    thread local storage (so one instance per request exists).

    At the end of the request the gathered data is logged (using signals sent
    by Django) using custom logger.
    """

    def __init__(self, get_response):
        """Initialize middleware."""
        self.get_response = get_response

    def _resolve_user(self, user):
        """Resolve user instance from request."""
        return get_anonymous_user() if user.is_anonymous else user

    def process_exception(self, request, exception):
        """Process the exception returned by processing the request.

        Log the exception and let other middleware handle it.
        """
        logger.error("Error processing request: '%s'.", exception)

    def __call__(self, request):
        """Process the request and store it.

        Also set the request id.
        """
        audit_manager = resolwe.auditlog.auditmanager.AuditManager.global_instance()
        audit_manager.reset()

        user = self._resolve_user(request.user)
        # Extract request id from the request headers or generate a random one.
        request_id_headers = ("HTTP_X_REQUEST_ID", "HTTP_X-AMZ-CF-ID")
        request_ids = [
            request.META.get(header)
            for header in request_id_headers
            if header in request.META
        ]
        request_id = request_ids[0] if request_ids else str(uuid.uuid4())
        request.META["RESOLWE_AUDIT_MANAGER_REQUEST_ID"] = request_id
        servername = getattr(request.META, "HTTP_HOST", "unknown")

        response = self.get_response(request)
        message = (
            "Request finished: "
            f"METHOD={request.method}; STATUS={response.status_code}; "
            f"USER={user.username}({user.id}); "
            f"URL={request.build_absolute_uri()}; "
            f"SERVERNAME={servername}"
        )

        audit_manager.log_message(message)
        audit_manager.emit(request, response)
        return response



# pylint: disable=missing-docstring
import logging
from time import time
from unittest.mock import MagicMock, patch

from django.core.exceptions import ValidationError
from rest_framework.response import Response

from resolwe.flow.managers.workload_connectors.kubernetes import (
    sanitize_kubernetes_label,
)
from resolwe.flow.models import Data, Process
from resolwe.flow.utils import get_data_checksum, retry
from resolwe.flow.utils.exceptions import resolwe_exception_handler
from resolwe.test import TestCase


class ExceptionsTestCase(TestCase):
    @patch("resolwe.flow.utils.exceptions.exception_handler")
    def test_exception_handler(self, exception_handler_mock):
        exception_handler_mock.return_value = None
        resp = resolwe_exception_handler(Exception, {})
        self.assertEqual(resp, None)

        exception_handler_mock.return_value = None
        resp = resolwe_exception_handler(ValidationError("Error description"), {})
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.data["error"], "Error description")

        original_resp = Response({})
        exception_handler_mock.return_value = original_resp
        resp = resolwe_exception_handler(ValidationError("Error description"), {})
        self.assertEqual(id(original_resp), id(resp))
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.data["error"], "Error description")

    def test_checksum_consistency(self):
        process = Process(version="1.0.0", slug="my-process")
        data = Data()

        data.input = {"tss": 0, "genome": "HG19"}
        checksum = get_data_checksum(data.input, process.slug, process.version)
        self.assertEqual(
            checksum, "ca322c2bb48b58eea3946e624fe6cfdc53c2cc12478465b6f0ca2d722e280c4c"
        )

        data.input = {"genome": "HG19", "tss": 0}
        checksum = get_data_checksum(data.input, process.slug, process.version)
        self.assertEqual(
            checksum, "ca322c2bb48b58eea3946e624fe6cfdc53c2cc12478465b6f0ca2d722e280c4c"
        )


class RetryTestCase(TestCase):
    """Test the retry decorator."""

    def test_retry_decorator(self):
        logger = logging.getLogger(__name__)
        callback = MagicMock()

        @retry(logger=logger, min_sleep=0.01, max_sleep=0.1, max_retries=3)
        def retry_ok():
            self.count += 1

        @retry(logger=logger, min_sleep=0.01, max_sleep=0.1, max_retries=3)
        def retry_one():
            self.count += 1
            if self.count < 2:
                raise Exception("Error")

        @retry(
            logger=logger,
            min_sleep=0.01,
            max_sleep=0.1,
            max_retries=3,
            cleanup_callback=callback,
        )
        def retry_one_callback():
            self.count += 1
            if self.count < 2:
                raise Exception("Error")

        @retry(logger=logger, min_sleep=0.01, max_sleep=0.1, max_retries=3)
        def retry_two():
            self.count += 1
            if self.count < 3:
                raise Exception("Error")

        @retry(logger=logger, min_sleep=0.01, max_sleep=0.07, max_retries=5)
        def retry_fail():
            self.count += 1
            raise Exception("Error")

        # Test method that does not raise exception.
        # Assert that:
        # - count is set to 1
        # - no message is logged
        self.count = 0
        with self.assertRaises(AssertionError):
            with self.assertLogs(logger):
                retry_ok()
        self.assertEqual(self.count, 1)

        # Test method that raises single exception.
        # Assert that:
        # - count is set to 2
        # - single message is logged
        self.count = 0
        start = time()
        with self.assertLogs(logger) as cm:
            retry_one()
        end = time()
        self.assertEqual(self.count, 2)
        self.assertGreater(end - start, 0.01)
        expected_logger_output = [
            "ERROR:resolwe.flow.tests.test_utils:Retry 1/3 got exception, will retry in 0.01 seconds.\nTraceback"
        ]
        self.assertEqual(len(cm.output), len(expected_logger_output))
        for i in range(len(expected_logger_output)):
            self.assertIn(expected_logger_output[i], cm.output[i])

        # Test method that raises single exception with cleanup callback.
        # Assert that:
        # - count is set to 2
        # - single message is logged
        # - callback was called once
        self.count = 0
        start = time()
        with self.assertLogs(logger) as cm:
            retry_one_callback()
        end = time()
        self.assertEqual(self.count, 2)
        self.assertGreater(end - start, 0.01)
        expected_logger_output = [
            "ERROR:resolwe.flow.tests.test_utils:Retry 1/3 got exception, will retry in 0.01 seconds.\nTraceback"
        ]
        self.assertEqual(len(cm.output), len(expected_logger_output))
        for i in range(len(expected_logger_output)):
            self.assertIn(expected_logger_output[i], cm.output[i])
        callback.assert_called_once()

        # Test method that raises two exceptions.
        # Assert that:
        # - count is set to 3
        # - two messages are logged
        # - the second sleep is 2 times the first.
        self.count = 0
        start = time()
        with self.assertLogs(logger) as cm:
            retry_two()
        end = time()
        self.assertEqual(self.count, 3)
        self.assertGreater(end - start, 0.03)
        expected_logger_output = [
            "ERROR:resolwe.flow.tests.test_utils:Retry 1/3 got exception, will retry in 0.01 seconds.\nTraceback",
            "ERROR:resolwe.flow.tests.test_utils:Retry 2/3 got exception, will retry in 0.02 seconds.\nTraceback",
        ]
        self.assertEqual(len(cm.output), len(expected_logger_output))
        for i in range(len(expected_logger_output)):
            self.assertIn(expected_logger_output[i], cm.output[i])

        # Test method that always raises exception. The following is checked:
        # - exception should be raised
        # - counter should be set to the max_retries
        # - last log message should be different
        # - sleeps should be exponential
        # - max_sleep must be upper limit for sleeps
        self.count = 0
        with self.assertLogs(logger) as cm:
            with self.assertRaises(Exception):
                start = time()
                retry_fail()
        end = time()
        self.assertGreater(end - start, 0.14)
        self.assertLess(end - start, 0.2 + 0.1)
        self.assertEqual(self.count, 5)
        expected_logger_output = [
            "ERROR:resolwe.flow.tests.test_utils:Retry 1/5 got exception, will retry in 0.01 seconds.\nTraceback",
            "ERROR:resolwe.flow.tests.test_utils:Retry 2/5 got exception, will retry in 0.02 seconds.\nTraceback",
            "ERROR:resolwe.flow.tests.test_utils:Retry 3/5 got exception, will retry in 0.04 seconds.\nTraceback",
            "ERROR:resolwe.flow.tests.test_utils:Retry 4/5 got exception, will retry in 0.07 seconds.\nTraceback",
            "ERROR:resolwe.flow.tests.test_utils:Retry 5/5 got exception, re-raising it.\nTraceback",
        ]
        self.assertEqual(len(cm.output), len(expected_logger_output))
        for i in range(len(expected_logger_output)):
            self.assertIn(expected_logger_output[i], cm.output[i])


class KubernetesTestCase(TestCase):
    def test_kubernetes_label_sanitizer(self):
        label = "this-is-a-valid-label"
        self.assertEqual(label, sanitize_kubernetes_label(label))

        too_long_label = "this-is-a-valid-label" * 10
        self.assertEqual(
            too_long_label[-63:], sanitize_kubernetes_label(too_long_label)
        )
        weird_label = "invalid/label with .some*weird'characters"
        self.assertEqual(
            "invalid-label-with-some-weird-characters",
            sanitize_kubernetes_label(weird_label),
        )

        label = "*/.2must start and end with alphanumeric character_"
        self.assertEqual(
            "2must-start-and-end-with-alphanumeric-character",
            sanitize_kubernetes_label(label),
        )

        label = "_" * 100 + "I am too long"
        self.assertEqual("I-am-too-long", sanitize_kubernetes_label(label))



# Generated by Django 3.2.12 on 2022-08-22 13:39

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0005_duplicate_data_dependency"),
    ]

    operations = [
        migrations.AddField(
            model_name="relation",
            name="descriptor",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="relation",
            name="descriptor_dirty",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="relation",
            name="descriptor_schema",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="flow.descriptorschema",
            ),
        ),
    ]



# Generated by Django 5.1.5 on 2025-02-23 10:11

from django.db import migrations, models


def migrate_annotations_values(apps, schema_editor):
    """Migrate annotation values to use deleted field.

    We have to:
    - set deleted field to True for all non-latest annotation values.
    - remove all delete markers (values with value=None).
    """
    AnnotationValue = apps.get_model("flow", "AnnotationValue")
    # Set deleted field to True for all non-latest annotation values.
    to_update_ids = list(
        AnnotationValue.objects.all()
        .annotate(
            rank=models.Window(
                expression=models.functions.DenseRank(),
                partition_by=[models.F("field"), models.F("entity")],
                order_by=models.F("created").desc(),
            ),
        )
        .filter(rank__gt=1)
        .values_list("id", flat=True)
    )
    AnnotationValue.objects.filter(id__in=to_update_ids).update(deleted=True)

    # Now remove all delete markers. This is done in couple steps:
    # 1. Save their ids.
    # 2. Set their values to something non-null (otherwise delete will fail).
    # 3. Delete them.
    to_delete_ids = AnnotationValue.objects.filter(
        models.Q(_value__isnull=True) | models.Q(_value__value__isnull=True)
    ).values_list("id", flat=True)
    AnnotationValue.objects.filter(id__in=to_delete_ids).update(
        _value={"value": "to_delete", "label": "to_delete"}
    )
    AnnotationValue.objects.filter(id__in=to_delete_ids).delete()


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0036_annotationvalue_deleted"),
    ]

    operations = [
        migrations.RunPython(migrate_annotations_values),
    ]



"""Permissions functions used in Resolwe Viewsets."""

from django.conf import settings
from django.db import models, transaction
from drf_spectacular.utils import extend_schema
from rest_framework import exceptions, serializers, status
from rest_framework.decorators import action
from rest_framework.request import Request
from rest_framework.response import Response

from resolwe.auditlog.auditmanager import AuditManager
from resolwe.flow.models.fields import VersionField
from resolwe.permissions.models import Permission
from resolwe.permissions.shortcuts import get_object_perms

from .utils import (
    check_owner_permission,
    check_public_permissions,
    check_user_permissions,
    update_permission,
)


class CurrentUserPermissionsSerializer(serializers.Serializer):
    """Current user permissions serializer."""

    id = serializers.IntegerField(required=False)
    type = serializers.ChoiceField(choices=["user", "group", "public"])
    name = serializers.CharField(max_length=100, required=False)
    username = serializers.CharField(max_length=150, required=False)
    permissions = serializers.ListField(
        child=serializers.ChoiceField(choices=list(map(str, Permission)))
    )


class PermissionSerializer(serializers.Serializer):
    """Permission serializer.

    Dictionary keys for users and groups are either ids or usernames (emails) or group names.
    """

    permission_choices = [Permission(i).name for i in list(Permission.highest())]

    users = serializers.DictField(
        child=serializers.ChoiceField(choices=permission_choices),
        required=False,
    )
    groups = serializers.DictField(
        child=serializers.ChoiceField(choices=permission_choices),
        required=False,
    )
    public = serializers.ChoiceField(choices=permission_choices, required=False)


class ResolwePermissionsMixin:
    """Mixin to support managing `Resolwe` objects' permissions."""

    def prefetch_current_user_permissions(self, queryset: models.QuerySet):
        """Prefetch permissions for the current user."""
        user = self.request.user
        filters = models.Q(user__username=settings.ANONYMOUS_USER_NAME)
        if not user.is_anonymous:
            filters |= models.Q(user=user) | models.Q(group__in=user.groups.all())

        qs_permission_model = self.qs_permission_model.filter(filters)
        return queryset.prefetch_related(
            models.Prefetch(
                "permission_group__permissions", queryset=qs_permission_model
            )
        )

    def get_serializer_class(self):
        """Augment base serializer class.

        Include permissions information with objects.

        """
        if not hasattr(type(self), "_permission_augmented_serializer_class"):
            augmented_serializer = self._augment_serializer_class()
            type(self)._permission_augmented_serializer_class = augmented_serializer

        return type(self)._permission_augmented_serializer_class

    def _augment_serializer_class(self):
        """Create an augmented serializer class."""
        base_class = super().get_serializer_class()

        class SerializerWithPermissions(base_class):
            """Augmented serializer."""

            serializer_field_mapping = {
                **base_class.serializer_field_mapping,
                VersionField: serializers.CharField,
            }

            def get_fields(serializer_self):
                """Return serializer's fields."""
                fields = super().get_fields()
                fields["current_user_permissions"] = CurrentUserPermissionsSerializer(
                    read_only=True, many=True
                )
                return fields

            def to_representation(serializer_self, instance: models.Model):
                """Object serializer."""
                data = super().to_representation(instance)
                if (
                    "fields" not in serializer_self.request.query_params
                    or "current_user_permissions"
                    in serializer_self.request.query_params["fields"]
                ):
                    to_serialize = get_object_perms(
                        instance, serializer_self.request.user, True
                    )
                    data["current_user_permissions"] = CurrentUserPermissionsSerializer(
                        to_serialize, read_only=True, many=True
                    ).data

                return data

        SerializerWithPermissions.__name__ = base_class.__name__ + "WithPermissions"
        SerializerWithPermissions.__doc__ = base_class.__doc__

        return SerializerWithPermissions

    @extend_schema(
        request=PermissionSerializer(),
        responses={status.HTTP_200_OK: CurrentUserPermissionsSerializer()},
        filters=False,
    )
    @action(
        detail=True,
        methods=["get", "post"],
        url_path="permissions",
        url_name="permissions",
    )
    def detail_permissions(self, request: Request, pk=None) -> Response:
        """Get or set permissions API endpoint."""
        # The object is taken from the queryset on the view for which
        # permissions are prefetched for the current user only.
        # This implies that obj.permission_group.permissions returns
        # permissions for the current user.
        # To get all the permissions we have to perform a refresh from the
        # database. This must only be done if user has share permission on the
        # given object otherwise only his permissions must be returned.
        obj = self.get_object()

        if obj.has_permission(Permission.SHARE, request.user):
            obj.refresh_from_db()

        audit_manager = AuditManager.global_instance()

        if request.method == "POST":
            audit_manager.log_message("Permissions updated: %s", request.data)
            allow_owner = obj.is_owner(request.user) or request.user.is_superuser
            check_owner_permission(request.data, allow_owner, obj)
            check_public_permissions(request.data)
            check_user_permissions(request.data, request.user.pk)

            with transaction.atomic():
                update_permission(obj, request.data)
                owner_count = obj.permission_group.permissions.filter(
                    value=Permission.OWNER.value, user__isnull=False
                ).count()

                if not owner_count:
                    raise exceptions.ParseError("Object must have at least one owner.")
        else:
            audit_manager.log_message("Permissions read: %s", request.data)

        return Response(get_object_perms(obj))



"""Custom database fields.

The VersionField is based on code at https://github.com/mcldev/django-versionfield .
"""

from django.conf import settings
from django.db import connection, models
from django.db.models import constants
from django.utils.text import slugify

from resolwe.flow.exceptions import SlugError

# Slug field settings.
MAX_SLUG_SEQUENCE_DIGITS = 9

# Version field settings.
DEFAULT_NUMBER_BITS = getattr(settings, "VERSION_FIELD_NUMBER_BITS", (8, 8, 16))
DEFAULT_INPUT_SIZE = getattr(settings, "VERSION_FIELD_WIDGET_INPUT_SIZE", 10)
DEFAULT_VERSION_VALUE = getattr(settings, "VERSION_FIELD_DEFAULT_VALUE", "0.0.0")


class ResolweSlugField(models.fields.SlugField):
    """Slug field."""

    def __init__(self, *args, **kwargs):
        """Initialize field."""
        self.populate_from = kwargs.pop("populate_from", None)

        # unique_with value can be string or tuple
        self.unique_with = kwargs.pop("unique_with", ())
        if isinstance(self.unique_with, str):
            self.unique_with = (self.unique_with,)

        if self.unique_with:
            # we will do "manual" granular check below
            kwargs["unique"] = False

        super().__init__(*args, **kwargs)

    def deconstruct(self):
        """Deconstruct method."""
        name, path, args, kwargs = super().deconstruct()

        if self.populate_from is not None:
            kwargs["populate_from"] = self.populate_from

        if self.unique_with != ():
            kwargs["unique_with"] = self.unique_with
            kwargs.pop("unique", None)

        return name, path, args, kwargs

    def _get_unique_constraints(self, instance):
        """Return SQL filter for filtering by fields in ``unique_with`` attribute.

        Filter is returned as tuple of two elements where first one is
        placeholder which is safe to insert into SQL query and second
        one may include potentially dangerous values and must be passed
        to SQL query in ``params`` attribute to make sure it is properly
        escaped.
        """
        constraints_expression = []
        constraints_values = {}
        for field_name in self.unique_with:
            if constants.LOOKUP_SEP in field_name:
                raise NotImplementedError(
                    "`unique_with` constraint does not support lookups by related models."
                )

            field = instance._meta.get_field(field_name)
            field_value = getattr(instance, field_name)

            # Convert value to the database representation.
            field_db_value = field.get_prep_value(field_value)

            constraint_key = "unique_" + field_name
            constraints_expression.append(
                "{} = %({})s".format(
                    connection.ops.quote_name(field.column), constraint_key
                )
            )
            constraints_values[constraint_key] = field_db_value

        if not constraints_expression:
            return "", []

        constraints_expression = "AND " + " AND ".join(constraints_expression)
        return constraints_expression, constraints_values

    def _get_populate_from_value(self, instance):
        """Get the value from ``populate_from`` attribute."""
        if hasattr(self.populate_from, "__call__"):
            # ResolweSlugField(populate_from=lambda instance: ...)
            return self.populate_from(instance)
        else:
            # ResolweSlugField(populate_from='foo')
            attr = getattr(instance, self.populate_from)
            return attr() if callable(attr) else attr

    def pre_save(self, instance, add):
        """Ensure slug uniqunes before save."""
        slug = self.value_from_object(instance)

        # We don't want to change slug defined by user.
        predefined_slug = bool(slug)

        if not slug and self.populate_from:
            slug = self._get_populate_from_value(instance)

        if original_slug := slug:
            slug = slugify(slug)
            if predefined_slug and original_slug != slug:
                raise SlugError(f"The value {original_slug} is not a valid slug.")

        if not slug:
            slug = None

            if not self.blank:
                slug = instance._meta.model_name
            elif not self.null:
                slug = ""

        if slug:
            # Make sure that auto generated slug with added sequence
            # won't excede maximal length.
            # Validation of predefined slugs is handled by Django.
            if not predefined_slug:
                slug = slug[: (self.max_length - MAX_SLUG_SEQUENCE_DIGITS - 1)]

            constraints_placeholder, constraints_values = self._get_unique_constraints(
                instance
            )

            instance_pk_name = instance._meta.pk.name

            # Safe values - make sure that there is no chance of SQL injection.
            query_params = {
                "constraints_placeholder": constraints_placeholder,
                "slug_column": connection.ops.quote_name(self.column),
                "slug_len": len(slug),
                "table_name": connection.ops.quote_name(self.model._meta.db_table),
                "pk_neq_placeholder": (
                    "AND {} != %(instance_pk)s".format(instance_pk_name)
                    if instance.pk
                    else ""
                ),
            }

            # SQL injection unsafe values - will be escaped.
            # Keys prefixed with `unique_` are reserved for `constraints_values` dict.
            query_escape_params = {
                "slug": slug,
                "slug_regex": "^{}(-[0-9]*)?$".format(slug),
            }
            query_escape_params.update(constraints_values)
            if instance.pk:
                query_escape_params["instance_pk"] = instance.pk

            with connection.cursor() as cursor:
                # TODO: Slowest part of this query is `MAX` function. It can
                #       be optimized by indexing slug column by slug sequence.
                #       https://www.postgresql.org/docs/9.4/static/indexes-expressional.html
                cursor.execute(
                    """
                    SELECT
                        CASE
                            WHEN (
                                EXISTS(
                                    SELECT 1 FROM {table_name} WHERE (
                                        {slug_column} = %(slug)s
                                        {pk_neq_placeholder}
                                        {constraints_placeholder}
                                    )
                                )
                            ) THEN MAX(slug_sequence) + 1
                            ELSE NULL
                        END
                    FROM (
                        SELECT COALESCE(
                            NULLIF(
                                RIGHT({slug_column}, -{slug_len}-1),
                                ''
                            ),
                            '1'
                        )::text::integer AS slug_sequence
                        FROM {table_name} WHERE (
                            {slug_column} ~ %(slug_regex)s
                            {pk_neq_placeholder}
                            {constraints_placeholder}
                        )
                    ) AS tmp
                    """.format(
                        **query_params
                    ),
                    params=query_escape_params,
                )
                result = cursor.fetchone()[0]

            if result is not None:
                if predefined_slug:
                    raise SlugError(
                        "Slug '{}' (version {}) is already taken.".format(
                            slug, instance.version
                        )
                    )

                if len(str(result)) > MAX_SLUG_SEQUENCE_DIGITS:
                    raise SlugError(
                        "Auto-generated slug sequence too long - please choose a different slug."
                    )

                slug = "{}-{}".format(slug, result)

            # Make the updated slug available as instance attribute.
            setattr(instance, self.name, slug)

        return slug


def convert_version_string_to_int(string: str, number_bits: list[int]) -> int:
    """Convert string version to int version.

    Take in a verison string e.g. '3.0.1'
    Store it as a converted int:
    3 * (2**number_bits[0]) + 0 * (2**number_bits[1]) + 1 * (2**number_bits[2])

    >>> convert_version_string_to_int('3.0.1',[8,8,16])
    50331649
    """
    # This is needed for using Version as a type in pydantic
    if string == Ellipsis:
        return string

    numbers = [int(number_string) for number_string in string.split(".")]

    if len(numbers) > len(number_bits):
        raise NotImplementedError(
            "Versions with more than {0} decimal places are not supported".format(
                len(number_bits) - 1
            )
        )

    # add 0s for missing numbers
    numbers.extend([0] * (len(number_bits) - len(numbers)))

    # convert to single int and return
    number = 0
    total_bits = 0
    for num, bits in reversed(list(zip(numbers, number_bits))):
        max_num = (bits + 1) - 1
        if num >= 1 << max_num:
            raise ValueError(
                "Number {0} cannot be stored with only {1} bits. Max is {2}".format(
                    num, bits, max_num
                )
            )
        number += num << total_bits
        total_bits += bits

    return number


def convert_version_int_to_string(number: int, number_bits: list[int]) -> str:
    """Convert int version to string version.

    >>> convert_version_int_to_string(50331649,[8,8,16])
    '3.0.1'
    """
    number_strings = []
    total_bits = sum(number_bits)
    for bits in number_bits:
        shift_amount = total_bits - bits
        number_segment = number >> shift_amount
        number_strings.append(str(number_segment))
        total_bits = total_bits - bits
        number = number - (number_segment << shift_amount)
    return ".".join(number_strings)


class Version:
    """The version object, used by the VersionField."""

    def __init__(self, string, number_bits=DEFAULT_NUMBER_BITS):
        """Create a version object from a string."""
        self.number_bits = number_bits
        self.internal_integer = convert_version_string_to_int(string, number_bits)

    def __str__(self):
        """Return a string representation."""
        return str(
            convert_version_int_to_string(self.internal_integer, self.number_bits)
        )

    def __repr__(self):
        """Return an internal string representation."""
        return self.__str__()

    def __int__(self):
        """Return the internal integer representation."""
        return self.internal_integer

    def __eq__(self, other):
        """Override equality operator."""
        if not other:
            return False  # we are obviously a valid Version, but 'other' isn't
        if other == Ellipsis:
            return False  # For pydantic use
        if isinstance(other, str):
            other = Version(other, self.number_bits)
        return int(self) == int(other)

    def __hash__(self):
        """Use the internal integer representation as hashing function."""
        return int(self)

    def __lt__(self, other):
        """Override less than operator."""
        if not other:
            return False
        if isinstance(other, str):
            other = Version(other, self.number_bits)
        return int(self) < int(other)

    def __le__(self, other):
        """Override less than or equal operator."""
        if not other:
            return False
        if isinstance(other, str):
            other = Version(other, self.number_bits)
        return int(self) <= int(other)

    def __gt__(self, other):
        """Override greater than operator."""
        if not other:
            return False
        if isinstance(other, str):
            other = Version(other, self.number_bits)
        return int(self) > int(other)

    def __ge__(self, other):
        """Override greater than or equal operator."""
        if not other:
            return False
        if isinstance(other, str):
            other = Version(other, self.number_bits)
        return int(self) >= int(other)

    def __add__(self, other):
        """Override add operator.

        For example: 1.0.1 + 2.0.1 = 3.0.2.
        """
        if not other:
            return self
        if isinstance(other, str):
            other = Version(other, self.number_bits)
        return Version(
            convert_version_int_to_string(int(self) + int(other), self.number_bits),
            self.number_bits,
        )

    def __sub__(self, other):
        """Override subtract operator.

        For example: 3.0.2 - 2.0.1 = 1.0.1. The minimal version is 0.0.0.
        """
        if not other:
            return self
        if isinstance(other, str):
            other = Version(other, self.number_bits)
        return Version(
            convert_version_int_to_string(
                max(int(self) - int(other), 0), self.number_bits
            ),
            self.number_bits,
        )


class VersionField(models.Field):
    """Store versions in field as strings.

    A Field where version numbers are input/output as strings (e.g. 3.0.1) but stored
    in the db as converted integers for fast indexing.
    """

    description = "A version number (e.g. 3.0.1)"

    def __init__(self, number_bits=DEFAULT_NUMBER_BITS, *args, **kwargs):
        """Initialize the field."""
        self.number_bits = number_bits
        super().__init__(*args, **kwargs)

    def db_type(self, connection):
        """Use integer as internal representation."""
        return "integer"

    def to_python(self, value):
        """Convert data to the instance of the Version object."""
        if not value:
            return None

        if isinstance(value, Version):
            return value

        if isinstance(value, str):
            return Version(value, self.number_bits)

        return Version(
            convert_version_int_to_string(value, self.number_bits), self.number_bits
        )

    def from_db_value(self, value, expression, connection):
        """Convert data from database representation."""
        if value is None:
            return value
        return Version(
            convert_version_int_to_string(value, self.number_bits), self.number_bits
        )

    def get_prep_value(self, value):
        """Convert data to the database representation."""
        if isinstance(value, str):
            return int(Version(value, self.number_bits))

        if value is None:
            return None

        return int(value)



#!/usr/bin/env python
import logging
import os
import sys

PROJECT_ROOT = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
sys.path.insert(0, PROJECT_ROOT)


if __name__ == "__main__":
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "tests.settings")

    # if len(sys.argv) > 1 and sys.argv[1] == 'runserver':
    #     raise ValueError('This Django project is not intended for running a server.')

    from django.core.management import execute_from_command_line

    execute_from_command_line(sys.argv)



# Generated by Django 5.1.5 on 2025-02-12 09:47

from django.db import migrations

import resolwe.flow.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0034_remove_data_idx_data_status_priority_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="annotationfield",
            name="version",
            field=resolwe.flow.models.fields.VersionField(default="0.0.0"),
        ),
        migrations.AlterField(
            model_name="annotationpreset",
            name="version",
            field=resolwe.flow.models.fields.VersionField(default="0.0.0"),
        ),
        migrations.AlterField(
            model_name="annotationvalue",
            name="version",
            field=resolwe.flow.models.fields.VersionField(default="0.0.0"),
        ),
        migrations.AlterField(
            model_name="collection",
            name="version",
            field=resolwe.flow.models.fields.VersionField(default="0.0.0"),
        ),
        migrations.AlterField(
            model_name="data",
            name="version",
            field=resolwe.flow.models.fields.VersionField(default="0.0.0"),
        ),
        migrations.AlterField(
            model_name="descriptorschema",
            name="version",
            field=resolwe.flow.models.fields.VersionField(default="0.0.0"),
        ),
        migrations.AlterField(
            model_name="entity",
            name="version",
            field=resolwe.flow.models.fields.VersionField(default="0.0.0"),
        ),
        migrations.AlterField(
            model_name="process",
            name="version",
            field=resolwe.flow.models.fields.VersionField(default="0.0.0"),
        ),
        migrations.AlterField(
            model_name="relation",
            name="version",
            field=resolwe.flow.models.fields.VersionField(default="0.0.0"),
        ),
        migrations.AlterField(
            model_name="storage",
            name="version",
            field=resolwe.flow.models.fields.VersionField(default="0.0.0"),
        ),
    ]



"""Storage viewset."""

from rest_framework import mixins, viewsets

from resolwe.flow.models import Storage
from resolwe.flow.serializers import StorageSerializer


class StorageViewSet(mixins.RetrieveModelMixin, viewsets.GenericViewSet):
    """API view for :class:`Storage` objects."""

    queryset = Storage.objects.all().select_related("contributor")
    serializer_class = StorageSerializer
